# Copyright (C) 2008-2011 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Tests for pack repositories.

These tests are repeated for all pack-based repository formats.
"""

from stat import S_ISDIR

from ... import controldir, errors, gpg, osutils, repository
from ... import revision as _mod_revision
from ... import tests, transport, ui
from ...tests import TestCaseWithTransport, TestNotApplicable, test_server
from ...transport import memory
from .. import inventory
from ..btree_index import BTreeGraphIndex
from ..groupcompress_repo import RepositoryFormat2a
from ..index import GraphIndex
from ..smart import client


class TestPackRepository(TestCaseWithTransport):
    """Tests to be repeated across all pack-based formats.

    The following are populated from the test scenario:

    :ivar format_name: Registered name fo the format to test.
    :ivar format_string: On-disk format marker.
    :ivar format_supports_external_lookups: Boolean.
    """

    def get_format(self):
        return controldir.format_registry.make_controldir(self.format_name)

    def test_attribute__fetch_order(self):
        """Packs do not need ordered data retrieval."""
        format = self.get_format()
        repo = self.make_repository('.', format=format)
        self.assertEqual('unordered', repo._format._fetch_order)

    def test_attribute__fetch_uses_deltas(self):
        """Packs reuse deltas."""
        format = self.get_format()
        repo = self.make_repository('.', format=format)
        if isinstance(format.repository_format, RepositoryFormat2a):
            # TODO: This is currently a workaround. CHK format repositories
            #       ignore the 'deltas' flag, but during conversions, we can't
            #       do unordered delta fetches. Remove this clause once we
            #       improve the inter-format fetching.
            self.assertEqual(False, repo._format._fetch_uses_deltas)
        else:
            self.assertEqual(True, repo._format._fetch_uses_deltas)

    def test_disk_layout(self):
        format = self.get_format()
        repo = self.make_repository('.', format=format)
        # in case of side effects of locking.
        repo.lock_write()
        repo.unlock()
        t = repo.controldir.get_repository_transport(None)
        self.check_format(t)
        # XXX: no locks left when unlocked at the moment
        # self.assertEqualDiff('', t.get('lock').read())
        self.check_databases(t)

    def check_format(self, t):
        with t.get('format') as f:
            self.assertEqualDiff(
                self.format_string.encode('ascii'),  # from scenario
                f.read())

    def assertHasNoKndx(self, t, knit_name):
        """Assert that knit_name has no index on t."""
        self.assertFalse(t.has(knit_name + '.kndx'))

    def assertHasNoKnit(self, t, knit_name):
        """Assert that knit_name exists on t."""
        # no default content
        self.assertFalse(t.has(knit_name + '.knit'))

    def check_databases(self, t):
        """check knit content for a repository."""
        # check conversion worked
        self.assertHasNoKndx(t, 'inventory')
        self.assertHasNoKnit(t, 'inventory')
        self.assertHasNoKndx(t, 'revisions')
        self.assertHasNoKnit(t, 'revisions')
        self.assertHasNoKndx(t, 'signatures')
        self.assertHasNoKnit(t, 'signatures')
        self.assertFalse(t.has('knits'))
        # revision-indexes file-container directory
        self.assertEqual([],
                         list(self.index_class(t, 'pack-names', None).iter_all_entries()))
        self.assertTrue(S_ISDIR(t.stat('packs').st_mode))
        self.assertTrue(S_ISDIR(t.stat('upload').st_mode))
        self.assertTrue(S_ISDIR(t.stat('indices').st_mode))
        self.assertTrue(S_ISDIR(t.stat('obsolete_packs').st_mode))

    def test_shared_disk_layout(self):
        format = self.get_format()
        repo = self.make_repository('.', shared=True, format=format)
        # we want:
        t = repo.controldir.get_repository_transport(None)
        self.check_format(t)
        # XXX: no locks left when unlocked at the moment
        # self.assertEqualDiff('', t.get('lock').read())
        # We should have a 'shared-storage' marker file.
        with t.get('shared-storage') as f:
            self.assertEqualDiff(b'', f.read())
        self.check_databases(t)

    def test_shared_no_tree_disk_layout(self):
        format = self.get_format()
        repo = self.make_repository('.', shared=True, format=format)
        repo.set_make_working_trees(False)
        # we want:
        t = repo.controldir.get_repository_transport(None)
        self.check_format(t)
        # XXX: no locks left when unlocked at the moment
        # self.assertEqualDiff('', t.get('lock').read())
        # We should have a 'shared-storage' marker file.
        with t.get('shared-storage') as f:
            self.assertEqualDiff(b'', f.read())
        # We should have a marker for the no-working-trees flag.
        with t.get('no-working-trees') as f:
            self.assertEqualDiff(b'', f.read())
        # The marker should go when we toggle the setting.
        repo.set_make_working_trees(True)
        self.assertFalse(t.has('no-working-trees'))
        self.check_databases(t)

    def test_adding_revision_creates_pack_indices(self):
        format = self.get_format()
        tree = self.make_branch_and_tree('.', format=format)
        trans = tree.branch.repository.controldir.get_repository_transport(
            None)
        self.assertEqual([],
                         list(self.index_class(trans, 'pack-names', None).iter_all_entries()))
        tree.commit('foobarbaz')
        index = self.index_class(trans, 'pack-names', None)
        index_nodes = list(index.iter_all_entries())
        self.assertEqual(1, len(index_nodes))
        node = index_nodes[0]
        name = node[1][0]
        # the pack sizes should be listed in the index
        pack_value = node[2]
        sizes = [int(digits) for digits in pack_value.split(b' ')]
        for size, suffix in zip(sizes, ['.rix', '.iix', '.tix', '.six']):
            stat = trans.stat('indices/{}{}'.format(name.decode('ascii'), suffix))
            self.assertEqual(size, stat.st_size)

    def test_pulling_nothing_leads_to_no_new_names(self):
        format = self.get_format()
        tree1 = self.make_branch_and_tree('1', format=format)
        tree2 = self.make_branch_and_tree('2', format=format)
        tree1.branch.repository.fetch(tree2.branch.repository)
        trans = tree1.branch.repository.controldir.get_repository_transport(
            None)
        self.assertEqual([],
                         list(self.index_class(trans, 'pack-names', None).iter_all_entries()))

    def test_commit_across_pack_shape_boundary_autopacks(self):
        format = self.get_format()
        tree = self.make_branch_and_tree('.', format=format)
        trans = tree.branch.repository.controldir.get_repository_transport(
            None)
        # This test could be a little cheaper by replacing the packs
        # attribute on the repository to allow a different pack distribution
        # and max packs policy - so we are checking the policy is honoured
        # in the test. But for now 11 commits is not a big deal in a single
        # test.
        for x in range(9):
            tree.commit('commit %s' % x)
        # there should be 9 packs:
        index = self.index_class(trans, 'pack-names', None)
        self.assertEqual(9, len(list(index.iter_all_entries())))
        # insert some files in obsolete_packs which should be removed by pack.
        trans.put_bytes('obsolete_packs/foo', b'123')
        trans.put_bytes('obsolete_packs/bar', b'321')
        # committing one more should coalesce to 1 of 10.
        tree.commit('commit triggering pack')
        index = self.index_class(trans, 'pack-names', None)
        self.assertEqual(1, len(list(index.iter_all_entries())))
        # packing should not damage data
        tree = tree.controldir.open_workingtree()
        check_result = tree.branch.repository.check(
            [tree.branch.last_revision()])
        nb_files = 5  # .pack, .rix, .iix, .tix, .six
        if tree.branch.repository._format.supports_chks:
            nb_files += 1  # .cix
        # We should have 10 x nb_files files in the obsolete_packs directory.
        obsolete_files = list(trans.list_dir('obsolete_packs'))
        self.assertFalse('foo' in obsolete_files)
        self.assertFalse('bar' in obsolete_files)
        self.assertEqual(10 * nb_files, len(obsolete_files))
        # XXX: Todo check packs obsoleted correctly - old packs and indices
        # in the obsolete_packs directory.
        large_pack_name = list(index.iter_all_entries())[0][1][0]
        # finally, committing again should not touch the large pack.
        tree.commit('commit not triggering pack')
        index = self.index_class(trans, 'pack-names', None)
        self.assertEqual(2, len(list(index.iter_all_entries())))
        pack_names = [node[1][0] for node in index.iter_all_entries()]
        self.assertTrue(large_pack_name in pack_names)

    def test_commit_write_group_returns_new_pack_names(self):
        # This test doesn't need real disk.
        self.vfs_transport_factory = memory.MemoryServer
        format = self.get_format()
        repo = self.make_repository('foo', format=format)
        with repo.lock_write():
            # All current pack repository styles autopack at 10 revisions; and
            # autopack as well as regular commit write group needs to return
            # the new pack name. Looping is a little ugly, but we don't have a
            # clean way to test both the autopack logic and the normal code
            # path without doing this loop.
            for pos in range(10):
                revid = b'%d' % pos
                repo.start_write_group()
                try:
                    inv = inventory.Inventory(revision_id=revid)
                    inv.root.revision = revid
                    repo.texts.add_lines((inv.root.file_id, revid), [], [])
                    rev = _mod_revision.Revision(timestamp=0, timezone=None,
                                                 committer="Foo Bar <foo@example.com>", message="Message",
                                                 revision_id=revid)
                    rev.parent_ids = ()
                    repo.add_revision(revid, rev, inv=inv)
                except:
                    repo.abort_write_group()
                    raise
                else:
                    old_names = set(repo._pack_collection._names)
                    result = repo.commit_write_group()
                    cur_names = set(repo._pack_collection._names)
                    # In this test, len(result) is always 1, so unordered is ok
                    new_names = list(cur_names - old_names)
                    self.assertEqual(new_names, result)

    def test_fail_obsolete_deletion(self):
        # failing to delete obsolete packs is not fatal
        format = self.get_format()
        server = test_server.FakeNFSServer()
        self.start_server(server)
        t = transport.get_transport_from_url(server.get_url())
        bzrdir = self.get_format().initialize_on_transport(t)
        repo = bzrdir.create_repository()
        repo_transport = bzrdir.get_repository_transport(None)
        self.assertTrue(repo_transport.has('obsolete_packs'))
        # these files are in use by another client and typically can't be deleted
        repo_transport.put_bytes('obsolete_packs/.nfsblahblah', b'contents')
        repo._pack_collection._clear_obsolete_packs()
        self.assertTrue(repo_transport.has('obsolete_packs/.nfsblahblah'))

    def test_pack_collection_sets_sibling_indices(self):
        """The CombinedGraphIndex objects in the pack collection are all
        siblings of each other, so that search-order reorderings will be copied
        to each other.
        """
        repo = self.make_repository('repo')
        pack_coll = repo._pack_collection
        indices = {pack_coll.revision_index, pack_coll.inventory_index,
                   pack_coll.text_index, pack_coll.signature_index}
        if pack_coll.chk_index is not None:
            indices.add(pack_coll.chk_index)
        combined_indices = {idx.combined_index for idx in indices}
        for combined_index in combined_indices:
            self.assertEqual(
                combined_indices.difference([combined_index]),
                combined_index._sibling_indices)

    def test_pack_with_signatures(self):
        format = self.get_format()
        tree = self.make_branch_and_tree('.', format=format)
        trans = tree.branch.repository.controldir.get_repository_transport(
            None)
        revid1 = tree.commit('start')
        revid2 = tree.commit('more work')
        strategy = gpg.LoopbackGPGStrategy(None)
        repo = tree.branch.repository
        self.addCleanup(repo.lock_write().unlock)
        repo.start_write_group()
        repo.sign_revision(revid1, strategy)
        repo.commit_write_group()
        repo.start_write_group()
        repo.sign_revision(revid2, strategy)
        repo.commit_write_group()
        tree.branch.repository.pack()
        # there should be 1 pack:
        index = self.index_class(trans, 'pack-names', None)
        self.assertEqual(1, len(list(index.iter_all_entries())))
        self.assertEqual(2, len(tree.branch.repository.all_revision_ids()))

    def test_pack_after_two_commits_packs_everything(self):
        format = self.get_format()
        tree = self.make_branch_and_tree('.', format=format)
        trans = tree.branch.repository.controldir.get_repository_transport(
            None)
        tree.commit('start')
        tree.commit('more work')
        tree.branch.repository.pack()
        # there should be 1 pack:
        index = self.index_class(trans, 'pack-names', None)
        self.assertEqual(1, len(list(index.iter_all_entries())))
        self.assertEqual(2, len(tree.branch.repository.all_revision_ids()))

    def test_pack_preserves_all_inventories(self):
        # This is related to bug:
        #   https://bugs.launchpad.net/bzr/+bug/412198
        # Stacked repositories need to keep the inventory for parents, even
        # after a pack operation. However, it is harder to test that, then just
        # test that all inventory texts are preserved.
        format = self.get_format()
        builder = self.make_branch_builder('source', format=format)
        builder.start_series()
        builder.build_snapshot(None, [
            ('add', ('', b'root-id', 'directory', None))],
            revision_id=b'A-id')
        builder.build_snapshot(None, [
            ('add', ('file', b'file-id', 'file', b'B content\n'))],
            revision_id=b'B-id')
        builder.build_snapshot(None, [
            ('modify', ('file', b'C content\n'))],
            revision_id=b'C-id')
        builder.finish_series()
        b = builder.get_branch()
        b.lock_read()
        self.addCleanup(b.unlock)
        repo = self.make_repository('repo', shared=True, format=format)
        repo.lock_write()
        self.addCleanup(repo.unlock)
        repo.fetch(b.repository, revision_id=b'B-id')
        inv = next(b.repository.iter_inventories([b'C-id']))
        repo.start_write_group()
        repo.add_inventory(b'C-id', inv, [b'B-id'])
        repo.commit_write_group()
        self.assertEqual([(b'A-id',), (b'B-id',), (b'C-id',)],
                         sorted(repo.inventories.keys()))
        repo.pack()
        self.assertEqual([(b'A-id',), (b'B-id',), (b'C-id',)],
                         sorted(repo.inventories.keys()))
        # Content should be preserved as well
        self.assertEqual(inv, next(repo.iter_inventories([b'C-id'])))

    def test_pack_layout(self):
        # Test that the ordering of revisions in pack repositories is
        # tip->ancestor
        format = self.get_format()
        tree = self.make_branch_and_tree('.', format=format)
        trans = tree.branch.repository.controldir.get_repository_transport(
            None)
        tree.commit('start', rev_id=b'1')
        tree.commit('more work', rev_id=b'2')
        tree.branch.repository.pack()
        tree.lock_read()
        self.addCleanup(tree.unlock)
        pack = tree.branch.repository._pack_collection.get_pack_by_name(
            tree.branch.repository._pack_collection.names()[0])
        # revision access tends to be tip->ancestor, so ordering that way on
        # disk is a good idea.
        pos_1 = pos_2 = None
        for _1, key, val, refs in pack.revision_index.iter_all_entries():
            if isinstance(format.repository_format, RepositoryFormat2a):
                # group_start, group_len, internal_start, internal_len
                pos = list(map(int, val.split()))
            else:
                # eol_flag, start, len
                pos = int(val[1:].split()[0])
            if key == (b'1',):
                pos_1 = pos
            else:
                pos_2 = pos
        self.assertTrue(pos_2 < pos_1, 'rev 1 came before rev 2 %s > %s'
                                       % (pos_1, pos_2))

    def test_pack_repositories_support_multiple_write_locks(self):
        format = self.get_format()
        self.make_repository('.', shared=True, format=format)
        r1 = repository.Repository.open('.')
        r2 = repository.Repository.open('.')
        r1.lock_write()
        self.addCleanup(r1.unlock)
        r2.lock_write()
        r2.unlock()

    def _add_text(self, repo, fileid):
        """Add a text to the repository within a write group."""
        repo.texts.add_lines((fileid, b'samplerev+' + fileid), [],
                             [b'smaplerev+' + fileid])

    def test_concurrent_writers_merge_new_packs(self):
        format = self.get_format()
        self.make_repository('.', shared=True, format=format)
        r1 = repository.Repository.open('.')
        r2 = repository.Repository.open('.')
        with r1.lock_write():
            # access enough data to load the names list
            list(r1.all_revision_ids())
            with r2.lock_write():
                # access enough data to load the names list
                list(r2.all_revision_ids())
                r1.start_write_group()
                try:
                    r2.start_write_group()
                    try:
                        self._add_text(r1, b'fileidr1')
                        self._add_text(r2, b'fileidr2')
                    except:
                        r2.abort_write_group()
                        raise
                except:
                    r1.abort_write_group()
                    raise
                # both r1 and r2 have open write groups with data in them
                # created while the other's write group was open.
                # Commit both which requires a merge to the pack-names.
                try:
                    r1.commit_write_group()
                except:
                    r1.abort_write_group()
                    r2.abort_write_group()
                    raise
                r2.commit_write_group()
                # tell r1 to reload from disk
                r1._pack_collection.reset()
                # Now both repositories should know about both names
                r1._pack_collection.ensure_loaded()
                r2._pack_collection.ensure_loaded()
                self.assertEqual(r1._pack_collection.names(),
                                 r2._pack_collection.names())
                self.assertEqual(2, len(r1._pack_collection.names()))

    def test_concurrent_writer_second_preserves_dropping_a_pack(self):
        format = self.get_format()
        self.make_repository('.', shared=True, format=format)
        r1 = repository.Repository.open('.')
        r2 = repository.Repository.open('.')
        # add a pack to drop
        with r1.lock_write():
            with repository.WriteGroup(r1):
                self._add_text(r1, b'fileidr1')
            r1._pack_collection.ensure_loaded()
            name_to_drop = r1._pack_collection.all_packs()[0].name
        with r1.lock_write():
            # access enough data to load the names list
            list(r1.all_revision_ids())
            with r2.lock_write():
                # access enough data to load the names list
                list(r2.all_revision_ids())
                r1._pack_collection.ensure_loaded()
                try:
                    r2.start_write_group()
                    try:
                        # in r1, drop the pack
                        r1._pack_collection._remove_pack_from_memory(
                            r1._pack_collection.get_pack_by_name(name_to_drop))
                        # in r2, add a pack
                        self._add_text(r2, b'fileidr2')
                    except:
                        r2.abort_write_group()
                        raise
                except:
                    r1._pack_collection.reset()
                    raise
                # r1 has a changed names list, and r2 an open write groups with
                # changes.
                # save r1, and then commit the r2 write group, which requires a
                # merge to the pack-names, which should not reinstate
                # name_to_drop
                try:
                    r1._pack_collection._save_pack_names()
                    r1._pack_collection.reset()
                except:
                    r2.abort_write_group()
                    raise
                try:
                    r2.commit_write_group()
                except:
                    r2.abort_write_group()
                    raise
                # Now both repositories should now about just one name.
                r1._pack_collection.ensure_loaded()
                r2._pack_collection.ensure_loaded()
                self.assertEqual(r1._pack_collection.names(),
                                 r2._pack_collection.names())
                self.assertEqual(1, len(r1._pack_collection.names()))
                self.assertFalse(name_to_drop in r1._pack_collection.names())

    def test_concurrent_pack_triggers_reload(self):
        # create 2 packs, which we will then collapse
        tree = self.make_branch_and_tree('tree')
        with tree.lock_write():
            rev1 = tree.commit('one')
            rev2 = tree.commit('two')
            r2 = repository.Repository.open('tree')
            with r2.lock_read():
                # Now r2 has read the pack-names file, but will need to reload
                # it after r1 has repacked
                tree.branch.repository.pack()
                self.assertEqual({rev2: (rev1,)}, r2.get_parent_map([rev2]))

    def test_concurrent_pack_during_get_record_reloads(self):
        tree = self.make_branch_and_tree('tree')
        with tree.lock_write():
            rev1 = tree.commit('one')
            rev2 = tree.commit('two')
            keys = [(rev1,), (rev2,)]
            r2 = repository.Repository.open('tree')
            with r2.lock_read():
                # At this point, we will start grabbing a record stream, and
                # trigger a repack mid-way
                packed = False
                result = {}
                record_stream = r2.revisions.get_record_stream(keys,
                                                               'unordered', False)
                for record in record_stream:
                    result[record.key] = record
                    if not packed:
                        tree.branch.repository.pack()
                        packed = True
                # The first record will be found in the original location, but
                # after the pack, we have to reload to find the next record
                self.assertEqual(sorted(keys), sorted(result.keys()))

    def test_concurrent_pack_during_autopack(self):
        tree = self.make_branch_and_tree('tree')
        with tree.lock_write():
            for i in range(9):
                tree.commit('rev %d' % (i,))
            r2 = repository.Repository.open('tree')
            with r2.lock_write():
                # Monkey patch so that pack occurs while the other repo is
                # autopacking. This is slightly bad, but all current pack
                # repository implementations have a _pack_collection, and we
                # test that it gets triggered. So if a future format changes
                # things, the test will fail rather than succeed accidentally.
                autopack_count = [0]
                r1 = tree.branch.repository
                orig = r1._pack_collection.pack_distribution

                def trigger_during_auto(*args, **kwargs):
                    ret = orig(*args, **kwargs)
                    if not autopack_count[0]:
                        r2.pack()
                    autopack_count[0] += 1
                    return ret
                r1._pack_collection.pack_distribution = trigger_during_auto
                tree.commit('autopack-rev')
                # This triggers 2 autopacks. The first one causes r2.pack() to
                # fire, but r2 doesn't see the new pack file yet. The
                # autopack restarts and sees there are 2 files and there
                # should be only 1 for 10 commits. So it goes ahead and
                # finishes autopacking.
                self.assertEqual([2], autopack_count)

    def test_lock_write_does_not_physically_lock(self):
        repo = self.make_repository('.', format=self.get_format())
        repo.lock_write()
        self.addCleanup(repo.unlock)
        self.assertFalse(repo.get_physical_lock_status())

    def prepare_for_break_lock(self):
        # Setup the global ui factory state so that a break-lock method call
        # will find usable input in the input stream.
        ui.ui_factory = ui.CannedInputUIFactory([True])

    def test_break_lock_breaks_physical_lock(self):
        repo = self.make_repository('.', format=self.get_format())
        repo._pack_collection.lock_names()
        repo.control_files.leave_in_place()
        repo.unlock()
        repo2 = repository.Repository.open('.')
        self.assertTrue(repo.get_physical_lock_status())
        self.prepare_for_break_lock()
        repo2.break_lock()
        self.assertFalse(repo.get_physical_lock_status())

    def test_broken_physical_locks_error_on__unlock_names_lock(self):
        repo = self.make_repository('.', format=self.get_format())
        repo._pack_collection.lock_names()
        self.assertTrue(repo.get_physical_lock_status())
        repo2 = repository.Repository.open('.')
        self.prepare_for_break_lock()
        repo2.break_lock()
        self.assertRaises(errors.LockBroken,
                          repo._pack_collection._unlock_names)

    def test_fetch_without_find_ghosts_ignores_ghosts(self):
        # we want two repositories at this point:
        # one with a revision that is a ghost in the other
        # repository.
        # 'ghost' is present in has_ghost, 'ghost' is absent in 'missing_ghost'.
        # 'references' is present in both repositories, and 'tip' is present
        # just in has_ghost.
        # has_ghost       missing_ghost
        # ------------------------------
        # 'ghost'             -
        # 'references'    'references'
        # 'tip'               -
        # In this test we fetch 'tip' which should not fetch 'ghost'
        has_ghost = self.make_repository('has_ghost', format=self.get_format())
        missing_ghost = self.make_repository('missing_ghost',
                                             format=self.get_format())

        def add_commit(repo, revision_id, parent_ids):
            repo.lock_write()
            repo.start_write_group()
            inv = inventory.Inventory(revision_id=revision_id)
            inv.root.revision = revision_id
            root_id = inv.root.file_id
            sha1 = repo.add_inventory(revision_id, inv, [])
            repo.texts.add_lines((root_id, revision_id), [], [])
            rev = _mod_revision.Revision(timestamp=0,
                                         timezone=None,
                                         committer="Foo Bar <foo@example.com>",
                                         message="Message",
                                         inventory_sha1=sha1,
                                         revision_id=revision_id)
            rev.parent_ids = parent_ids
            repo.add_revision(revision_id, rev)
            repo.commit_write_group()
            repo.unlock()
        add_commit(has_ghost, b'ghost', [])
        add_commit(has_ghost, b'references', [b'ghost'])
        add_commit(missing_ghost, b'references', [b'ghost'])
        add_commit(has_ghost, b'tip', [b'references'])
        missing_ghost.fetch(has_ghost, b'tip')
        # missing ghost now has tip and not ghost.
        rev = missing_ghost.get_revision(b'tip')
        inv = missing_ghost.get_inventory(b'tip')
        self.assertRaises(errors.NoSuchRevision,
                          missing_ghost.get_revision, b'ghost')
        self.assertRaises(errors.NoSuchRevision,
                          missing_ghost.get_inventory, b'ghost')

    def make_write_ready_repo(self):
        format = self.get_format()
        if isinstance(format.repository_format, RepositoryFormat2a):
            raise TestNotApplicable("No missing compression parents")
        repo = self.make_repository('.', format=format)
        repo.lock_write()
        self.addCleanup(repo.unlock)
        repo.start_write_group()
        self.addCleanup(repo.abort_write_group)
        return repo

    def test_missing_inventories_compression_parent_prevents_commit(self):
        repo = self.make_write_ready_repo()
        key = ('junk',)
        repo.inventories._index._missing_compression_parents.add(key)
        self.assertRaises(errors.BzrCheckError, repo.commit_write_group)
        self.assertRaises(errors.BzrCheckError, repo.commit_write_group)

    def test_missing_revisions_compression_parent_prevents_commit(self):
        repo = self.make_write_ready_repo()
        key = ('junk',)
        repo.revisions._index._missing_compression_parents.add(key)
        self.assertRaises(errors.BzrCheckError, repo.commit_write_group)
        self.assertRaises(errors.BzrCheckError, repo.commit_write_group)

    def test_missing_signatures_compression_parent_prevents_commit(self):
        repo = self.make_write_ready_repo()
        key = ('junk',)
        repo.signatures._index._missing_compression_parents.add(key)
        self.assertRaises(errors.BzrCheckError, repo.commit_write_group)
        self.assertRaises(errors.BzrCheckError, repo.commit_write_group)

    def test_missing_text_compression_parent_prevents_commit(self):
        repo = self.make_write_ready_repo()
        key = ('some', 'junk')
        repo.texts._index._missing_compression_parents.add(key)
        self.assertRaises(errors.BzrCheckError, repo.commit_write_group)
        e = self.assertRaises(errors.BzrCheckError, repo.commit_write_group)

    def test_supports_external_lookups(self):
        repo = self.make_repository('.', format=self.get_format())
        self.assertEqual(self.format_supports_external_lookups,
                         repo._format.supports_external_lookups)

    def _lock_write(self, write_lockable):
        """Lock write_lockable, add a cleanup and return the result.

        :param write_lockable: An object with a lock_write method.
        :return: The result of write_lockable.lock_write().
        """
        result = write_lockable.lock_write()
        self.addCleanup(result.unlock)
        return result

    def test_abort_write_group_does_not_raise_when_suppressed(self):
        """Similar to per_repository.test_write_group's test of the same name.

        Also requires that the exception is logged.
        """
        self.vfs_transport_factory = memory.MemoryServer
        repo = self.make_repository('repo', format=self.get_format())
        token = self._lock_write(repo).repository_token
        repo.start_write_group()
        # Damage the repository on the filesystem
        self.get_transport('').rename('repo', 'foo')
        # abort_write_group will not raise an error
        self.assertEqual(None, repo.abort_write_group(suppress_errors=True))
        # But it does log an error
        log = self.get_log()
        self.assertContainsRe(log, 'abort_write_group failed')
        self.assertContainsRe(log, r'INFO  brz: ERROR \(ignored\):')
        if token is not None:
            repo.leave_lock_in_place()

    def test_abort_write_group_does_raise_when_not_suppressed(self):
        self.vfs_transport_factory = memory.MemoryServer
        repo = self.make_repository('repo', format=self.get_format())
        token = self._lock_write(repo).repository_token
        repo.start_write_group()
        # Damage the repository on the filesystem
        self.get_transport('').rename('repo', 'foo')
        # abort_write_group will not raise an error
        self.assertRaises(Exception, repo.abort_write_group)
        if token is not None:
            repo.leave_lock_in_place()

    def test_suspend_write_group(self):
        self.vfs_transport_factory = memory.MemoryServer
        repo = self.make_repository('repo', format=self.get_format())
        token = self._lock_write(repo).repository_token
        repo.start_write_group()
        repo.texts.add_lines((b'file-id', b'revid'), (), [b'lines'])
        wg_tokens = repo.suspend_write_group()
        expected_pack_name = wg_tokens[0] + '.pack'
        expected_names = [wg_tokens[0] + ext for ext in
                          ('.rix', '.iix', '.tix', '.six')]
        if repo.chk_bytes is not None:
            expected_names.append(wg_tokens[0] + '.cix')
        expected_names.append(expected_pack_name)
        upload_transport = repo._pack_collection._upload_transport
        limbo_files = upload_transport.list_dir('')
        self.assertEqual(sorted(expected_names), sorted(limbo_files))
        md5 = osutils.md5(upload_transport.get_bytes(expected_pack_name))
        self.assertEqual(wg_tokens[0], md5.hexdigest())

    def test_resume_chk_bytes(self):
        self.vfs_transport_factory = memory.MemoryServer
        repo = self.make_repository('repo', format=self.get_format())
        if repo.chk_bytes is None:
            raise TestNotApplicable('no chk_bytes for this repository')
        token = self._lock_write(repo).repository_token
        repo.start_write_group()
        text = b'a bit of text\n'
        key = (b'sha1:' + osutils.sha_string(text),)
        repo.chk_bytes.add_lines(key, (), [text])
        wg_tokens = repo.suspend_write_group()
        same_repo = repo.controldir.open_repository()
        same_repo.lock_write()
        self.addCleanup(same_repo.unlock)
        same_repo.resume_write_group(wg_tokens)
        self.assertEqual([key], list(same_repo.chk_bytes.keys()))
        self.assertEqual(
            text, next(same_repo.chk_bytes.get_record_stream(
                [key], 'unordered', True)).get_bytes_as('fulltext'))
        same_repo.abort_write_group()
        self.assertEqual([], list(same_repo.chk_bytes.keys()))

    def test_resume_write_group_then_abort(self):
        # Create a repo, start a write group, insert some data, suspend.
        self.vfs_transport_factory = memory.MemoryServer
        repo = self.make_repository('repo', format=self.get_format())
        token = self._lock_write(repo).repository_token
        repo.start_write_group()
        text_key = (b'file-id', b'revid')
        repo.texts.add_lines(text_key, (), [b'lines'])
        wg_tokens = repo.suspend_write_group()
        # Get a fresh repository object for the repo on the filesystem.
        same_repo = repo.controldir.open_repository()
        # Resume
        same_repo.lock_write()
        self.addCleanup(same_repo.unlock)
        same_repo.resume_write_group(wg_tokens)
        same_repo.abort_write_group()
        self.assertEqual(
            [], same_repo._pack_collection._upload_transport.list_dir(''))
        self.assertEqual(
            [], same_repo._pack_collection._pack_transport.list_dir(''))

    def test_commit_resumed_write_group(self):
        self.vfs_transport_factory = memory.MemoryServer
        repo = self.make_repository('repo', format=self.get_format())
        token = self._lock_write(repo).repository_token
        repo.start_write_group()
        text_key = (b'file-id', b'revid')
        repo.texts.add_lines(text_key, (), [b'lines'])
        wg_tokens = repo.suspend_write_group()
        # Get a fresh repository object for the repo on the filesystem.
        same_repo = repo.controldir.open_repository()
        # Resume
        same_repo.lock_write()
        self.addCleanup(same_repo.unlock)
        same_repo.resume_write_group(wg_tokens)
        same_repo.commit_write_group()
        expected_pack_name = wg_tokens[0] + '.pack'
        expected_names = [wg_tokens[0] + ext for ext in
                          ('.rix', '.iix', '.tix', '.six')]
        if repo.chk_bytes is not None:
            expected_names.append(wg_tokens[0] + '.cix')
        self.assertEqual(
            [], same_repo._pack_collection._upload_transport.list_dir(''))
        index_names = repo._pack_collection._index_transport.list_dir('')
        self.assertEqual(sorted(expected_names), sorted(index_names))
        pack_names = repo._pack_collection._pack_transport.list_dir('')
        self.assertEqual([expected_pack_name], pack_names)

    def test_resume_malformed_token(self):
        self.vfs_transport_factory = memory.MemoryServer
        # Make a repository with a suspended write group
        repo = self.make_repository('repo', format=self.get_format())
        token = self._lock_write(repo).repository_token
        repo.start_write_group()
        text_key = (b'file-id', b'revid')
        repo.texts.add_lines(text_key, (), [b'lines'])
        wg_tokens = repo.suspend_write_group()
        # Make a new repository
        new_repo = self.make_repository('new_repo', format=self.get_format())
        token = self._lock_write(new_repo).repository_token
        hacked_wg_token = (
            '../../../../repo/.bzr/repository/upload/' + wg_tokens[0])
        self.assertRaises(
            errors.UnresumableWriteGroup,
            new_repo.resume_write_group, [hacked_wg_token])


class TestPackRepositoryStacking(TestCaseWithTransport):

    """Tests for stacking pack repositories"""

    def setUp(self):
        if not self.format_supports_external_lookups:
            raise TestNotApplicable("%r doesn't support stacking"
                                    % (self.format_name,))
        super().setUp()

    def get_format(self):
        return controldir.format_registry.make_controldir(self.format_name)

    def test_stack_checks_rich_root_compatibility(self):
        # early versions of the packing code relied on pack internals to
        # stack, but the current version should be able to stack on any
        # format.
        #
        # TODO: Possibly this should be run per-repository-format and raise
        # TestNotApplicable on formats that don't support stacking. -- mbp
        # 20080729
        repo = self.make_repository('repo', format=self.get_format())
        if repo.supports_rich_root():
            # can only stack on repositories that have compatible internal
            # metadata
            if getattr(repo._format, 'supports_tree_reference', False):
                matching_format_name = '2a'
            else:
                if repo._format.supports_chks:
                    matching_format_name = '2a'
                else:
                    matching_format_name = 'rich-root-pack'
            mismatching_format_name = 'pack-0.92'
        else:
            # We don't have a non-rich-root CHK format.
            if repo._format.supports_chks:
                raise AssertionError("no non-rich-root CHK formats known")
            else:
                matching_format_name = 'pack-0.92'
            mismatching_format_name = 'pack-0.92-subtree'
        base = self.make_repository('base', format=matching_format_name)
        repo.add_fallback_repository(base)
        # you can't stack on something with incompatible data
        bad_repo = self.make_repository('mismatch',
                                        format=mismatching_format_name)
        e = self.assertRaises(errors.IncompatibleRepositories,
                              repo.add_fallback_repository, bad_repo)
        self.assertContainsRe(str(e),
                              r'(?m)KnitPackRepository.*/mismatch/.*\nis not compatible with\n'
                              r'.*Repository.*/repo/.*\n'
                              r'different rich-root support')

    def test_stack_checks_serializers_compatibility(self):
        repo = self.make_repository('repo', format=self.get_format())
        if getattr(repo._format, 'supports_tree_reference', False):
            # can only stack on repositories that have compatible internal
            # metadata
            matching_format_name = '2a'
            mismatching_format_name = 'rich-root-pack'
        else:
            if repo.supports_rich_root():
                if repo._format.supports_chks:
                    matching_format_name = '2a'
                else:
                    matching_format_name = 'rich-root-pack'
                mismatching_format_name = 'pack-0.92-subtree'
            else:
                raise TestNotApplicable('No formats use non-v5 serializer'
                                        ' without having rich-root also set')
        base = self.make_repository('base', format=matching_format_name)
        repo.add_fallback_repository(base)
        # you can't stack on something with incompatible data
        bad_repo = self.make_repository('mismatch',
                                        format=mismatching_format_name)
        e = self.assertRaises(errors.IncompatibleRepositories,
                              repo.add_fallback_repository, bad_repo)
        self.assertContainsRe(str(e),
                              r'(?m)KnitPackRepository.*/mismatch/.*\nis not compatible with\n'
                              r'.*Repository.*/repo/.*\n'
                              r'different serializers')

    def test_adding_pack_does_not_record_pack_names_from_other_repositories(self):
        base = self.make_branch_and_tree('base', format=self.get_format())
        base.commit('foo')
        referencing = self.make_branch_and_tree(
            'repo', format=self.get_format())
        referencing.branch.repository.add_fallback_repository(
            base.branch.repository)
        local_tree = referencing.branch.create_checkout('local')
        local_tree.commit('bar')
        new_instance = referencing.controldir.open_repository()
        new_instance.lock_read()
        self.addCleanup(new_instance.unlock)
        new_instance._pack_collection.ensure_loaded()
        self.assertEqual(1, len(new_instance._pack_collection.all_packs()))

    def test_autopack_only_considers_main_repo_packs(self):
        format = self.get_format()
        base = self.make_branch_and_tree('base', format=format)
        base.commit('foo')
        tree = self.make_branch_and_tree('repo', format=format)
        tree.branch.repository.add_fallback_repository(base.branch.repository)
        trans = tree.branch.repository.controldir.get_repository_transport(
            None)
        # This test could be a little cheaper by replacing the packs
        # attribute on the repository to allow a different pack distribution
        # and max packs policy - so we are checking the policy is honoured
        # in the test. But for now 11 commits is not a big deal in a single
        # test.
        local_tree = tree.branch.create_checkout('local')
        for x in range(9):
            local_tree.commit('commit %s' % x)
        # there should be 9 packs:
        index = self.index_class(trans, 'pack-names', None)
        self.assertEqual(9, len(list(index.iter_all_entries())))
        # committing one more should coalesce to 1 of 10.
        local_tree.commit('commit triggering pack')
        index = self.index_class(trans, 'pack-names', None)
        self.assertEqual(1, len(list(index.iter_all_entries())))
        # packing should not damage data
        tree = tree.controldir.open_workingtree()
        check_result = tree.branch.repository.check(
            [tree.branch.last_revision()])
        nb_files = 5  # .pack, .rix, .iix, .tix, .six
        if tree.branch.repository._format.supports_chks:
            nb_files += 1  # .cix
        # We should have 10 x nb_files files in the obsolete_packs directory.
        obsolete_files = list(trans.list_dir('obsolete_packs'))
        self.assertFalse('foo' in obsolete_files)
        self.assertFalse('bar' in obsolete_files)
        self.assertEqual(10 * nb_files, len(obsolete_files))
        # XXX: Todo check packs obsoleted correctly - old packs and indices
        # in the obsolete_packs directory.
        large_pack_name = list(index.iter_all_entries())[0][1][0]
        # finally, committing again should not touch the large pack.
        local_tree.commit('commit not triggering pack')
        index = self.index_class(trans, 'pack-names', None)
        self.assertEqual(2, len(list(index.iter_all_entries())))
        pack_names = [node[1][0] for node in index.iter_all_entries()]
        self.assertTrue(large_pack_name in pack_names)


class TestKeyDependencies(TestCaseWithTransport):

    def get_format(self):
        return controldir.format_registry.make_controldir(self.format_name)

    def create_source_and_target(self):
        builder = self.make_branch_builder('source', format=self.get_format())
        builder.start_series()
        builder.build_snapshot(None, [
            ('add', ('', b'root-id', 'directory', None))],
            revision_id=b'A-id')
        builder.build_snapshot(
            [b'A-id', b'ghost-id'], [],
            revision_id=b'B-id', )
        builder.finish_series()
        repo = self.make_repository('target', format=self.get_format())
        b = builder.get_branch()
        b.lock_read()
        self.addCleanup(b.unlock)
        repo.lock_write()
        self.addCleanup(repo.unlock)
        return b.repository, repo

    def test_key_dependencies_cleared_on_abort(self):
        source_repo, target_repo = self.create_source_and_target()
        target_repo.start_write_group()
        try:
            stream = source_repo.revisions.get_record_stream([(b'B-id',)],
                                                             'unordered', True)
            target_repo.revisions.insert_record_stream(stream)
            key_refs = target_repo.revisions._index._key_dependencies
            self.assertEqual([(b'B-id',)], sorted(key_refs.get_referrers()))
        finally:
            target_repo.abort_write_group()
        self.assertEqual([], sorted(key_refs.get_referrers()))

    def test_key_dependencies_cleared_on_suspend(self):
        source_repo, target_repo = self.create_source_and_target()
        target_repo.start_write_group()
        try:
            stream = source_repo.revisions.get_record_stream([(b'B-id',)],
                                                             'unordered', True)
            target_repo.revisions.insert_record_stream(stream)
            key_refs = target_repo.revisions._index._key_dependencies
            self.assertEqual([(b'B-id',)], sorted(key_refs.get_referrers()))
        finally:
            target_repo.suspend_write_group()
        self.assertEqual([], sorted(key_refs.get_referrers()))

    def test_key_dependencies_cleared_on_commit(self):
        source_repo, target_repo = self.create_source_and_target()
        target_repo.start_write_group()
        try:
            # Copy all texts, inventories, and chks so that nothing is missing
            # for revision B-id.
            for vf_name in ['texts', 'chk_bytes', 'inventories']:
                source_vf = getattr(source_repo, vf_name, None)
                if source_vf is None:
                    continue
                target_vf = getattr(target_repo, vf_name)
                stream = source_vf.get_record_stream(
                    source_vf.keys(), 'unordered', True)
                target_vf.insert_record_stream(stream)
            # Copy just revision B-id
            stream = source_repo.revisions.get_record_stream(
                [(b'B-id',)], 'unordered', True)
            target_repo.revisions.insert_record_stream(stream)
            key_refs = target_repo.revisions._index._key_dependencies
            self.assertEqual([(b'B-id',)], sorted(key_refs.get_referrers()))
        finally:
            target_repo.commit_write_group()
        self.assertEqual([], sorted(key_refs.get_referrers()))


class TestSmartServerAutopack(TestCaseWithTransport):

    def setUp(self):
        super().setUp()
        # Create a smart server that publishes whatever the backing VFS server
        # does.
        self.smart_server = test_server.SmartTCPServer_for_testing()
        self.start_server(self.smart_server, self.get_server())
        # Log all HPSS calls into self.hpss_calls.
        client._SmartClient.hooks.install_named_hook(
            'call', self.capture_hpss_call, None)
        self.hpss_calls = []

    def capture_hpss_call(self, params):
        self.hpss_calls.append(params.method)

    def get_format(self):
        return controldir.format_registry.make_controldir(self.format_name)

    def test_autopack_or_streaming_rpc_is_used_when_using_hpss(self):
        # Make local and remote repos
        format = self.get_format()
        tree = self.make_branch_and_tree('local', format=format)
        self.make_branch_and_tree('remote', format=format)
        remote_branch_url = self.smart_server.get_url() + 'remote'
        remote_branch = controldir.ControlDir.open(
            remote_branch_url).open_branch()
        # Make 9 local revisions, and push them one at a time to the remote
        # repo to produce 9 pack files.
        for x in range(9):
            tree.commit('commit %s' % x)
            tree.branch.push(remote_branch)
        # Make one more push to trigger an autopack
        self.hpss_calls = []
        tree.commit('commit triggering pack')
        tree.branch.push(remote_branch)
        autopack_calls = len([call for call in self.hpss_calls if call
                              == b'PackRepository.autopack'])
        streaming_calls = len([call for call in self.hpss_calls if call in
                               (b'Repository.insert_stream', b'Repository.insert_stream_1.19')])
        if autopack_calls:
            # Non streaming server
            self.assertEqual(1, autopack_calls)
            self.assertEqual(0, streaming_calls)
        else:
            # Streaming was used, which autopacks on the remote end.
            self.assertEqual(0, autopack_calls)
            # NB: The 2 calls are because of the sanity check that the server
            # supports the verb (see remote.py:RemoteSink.insert_stream for
            # details).
            self.assertEqual(2, streaming_calls)


def load_tests(loader, basic_tests, pattern):
    # these give the bzrdir canned format name, and the repository on-disk
    # format string
    scenarios_params = [
        dict(format_name='pack-0.92',
             format_string="Bazaar pack repository format 1 (needs bzr 0.92)\n",
             format_supports_external_lookups=False,
             index_class=GraphIndex),
        dict(format_name='pack-0.92-subtree',
             format_string="Bazaar pack repository format 1 "
             "with subtree support (needs bzr 0.92)\n",
             format_supports_external_lookups=False,
             index_class=GraphIndex),
        dict(format_name='1.6',
             format_string="Bazaar RepositoryFormatKnitPack5 (bzr 1.6)\n",
             format_supports_external_lookups=True,
             index_class=GraphIndex),
        dict(format_name='1.6.1-rich-root',
             format_string="Bazaar RepositoryFormatKnitPack5RichRoot "
             "(bzr 1.6.1)\n",
             format_supports_external_lookups=True,
             index_class=GraphIndex),
        dict(format_name='1.9',
             format_string="Bazaar RepositoryFormatKnitPack6 (bzr 1.9)\n",
             format_supports_external_lookups=True,
             index_class=BTreeGraphIndex),
        dict(format_name='1.9-rich-root',
             format_string="Bazaar RepositoryFormatKnitPack6RichRoot "
             "(bzr 1.9)\n",
             format_supports_external_lookups=True,
             index_class=BTreeGraphIndex),
        dict(format_name='2a',
             format_string="Bazaar repository format 2a "
             "(needs bzr 1.16 or later)\n",
             format_supports_external_lookups=True,
             index_class=BTreeGraphIndex),
        ]
    # name of the scenario is the format name
    scenarios = [(s['format_name'], s) for s in scenarios_params]
    return tests.multiply_tests(basic_tests, scenarios, loader.suiteClass())



# Copyright (C) 2005-2011 Canonical Ltd
#       Author: Robert Collins <robert.collins@canonical.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
#

import logging
import unittest
import weakref
from typing import Dict, List

from .. import pyutils

# Mark this python module as being part of the implementation
# of unittest: this gives us better tracebacks where the last
# shown frame is the test code, not our assertXYZ.
__unittest = 1


class LogCollector(logging.Handler):

    def __init__(self):
        logging.Handler.__init__(self)
        self.records = []

    def emit(self, record):
        self.records.append(record.getMessage())


def makeCollectingLogger():
    """I make a logger instance that collects its logs for programmatic analysis
    -> (logger, collector)"""
    logger = logging.Logger("collector")
    handler = LogCollector()
    handler.setFormatter(logging.Formatter("%(levelname)s: %(message)s"))
    logger.addHandler(handler)
    return logger, handler


def visitTests(suite, visitor):
    """A foreign method for visiting the tests in a test suite."""
    for test in suite._tests:
        # Abusing types to avoid monkey patching unittest.TestCase.
        # Maybe that would be better?
        try:
            test.visit(visitor)
        except AttributeError:
            if isinstance(test, unittest.TestCase):
                visitor.visitCase(test)
            elif isinstance(test, unittest.TestSuite):
                visitor.visitSuite(test)
                visitTests(test, visitor)
            else:
                print("unvisitable non-unittest.TestCase element {!r} ({!r})".format(
                    test, test.__class__))


class FailedCollectionCase(unittest.TestCase):
    """Pseudo-test to run and report failure if given case was uncollected"""

    def __init__(self, case):
        super().__init__("fail_uncollected")
        # GZ 2011-09-16: Maybe catch errors from id() method as cases may be
        #                in a bit of a funny state by now.
        self._problem_case_id = case.id()

    def id(self):
        if self._problem_case_id[-1:] == ")":
            return self._problem_case_id[:-1] + ",uncollected)"
        return self._problem_case_id + "(uncollected)"

    def fail_uncollected(self):
        self.fail("Uncollected test case: " + self._problem_case_id)


class TestSuite(unittest.TestSuite):
    """I am an extended TestSuite with a visitor interface.
    This is primarily to allow filtering of tests - and suites or
    more in the future. An iterator of just tests wouldn't scale..."""

    def visit(self, visitor):
        """visit the composite. Visiting is depth-first.
        current callbacks are visitSuite and visitCase."""
        visitor.visitSuite(self)
        visitTests(self, visitor)

    def run(self, result):
        """Run the tests in the suite, discarding references after running."""
        tests = list(self)
        tests.reverse()
        self._tests = []
        stored_count = 0
        count_stored_tests = getattr(result, "_count_stored_tests", int)
        from breezy.tests import selftest_debug_flags
        notify = "uncollected_cases" in selftest_debug_flags
        while tests:
            if result.shouldStop:
                self._tests = reversed(tests)
                break
            case = _run_and_collect_case(tests.pop(), result)()
            new_stored_count = count_stored_tests()
            if case is not None and isinstance(case, unittest.TestCase):
                if stored_count == new_stored_count and notify:
                    # Testcase didn't fail, but somehow is still alive
                    FailedCollectionCase(case).run(result)
                    # Adding a new failure so need to reupdate the count
                    new_stored_count = count_stored_tests()
                # GZ 2011-09-16: Previously zombied the case at this point by
                #                clearing the dict as fallback, skip for now.
            stored_count = new_stored_count
        return result


def _run_and_collect_case(case, res):
    """Run test case against result and use weakref to drop the refcount"""
    case.run(res)
    return weakref.ref(case)


class TestLoader(unittest.TestLoader):
    """Custom TestLoader to extend the stock python one."""

    suiteClass = TestSuite
    # Memoize test names by test class dict
    test_func_names: Dict[str, List[str]] = {}

    def loadTestsFromModuleNames(self, names):
        """use a custom means to load tests from modules.

        There is an undesirable glitch in the python TestLoader where a
        import error is ignore. We think this can be solved by ensuring the
        requested name is resolvable, if its not raising the original error.
        """
        result = self.suiteClass()
        for name in names:
            result.addTests(self.loadTestsFromModuleName(name))
        return result

    def loadTestsFromModuleName(self, name):
        result = self.suiteClass()
        module = pyutils.get_named_object(name)

        result.addTests(self.loadTestsFromModule(module))
        return result

    def getTestCaseNames(self, test_case_class):
        test_fn_names = self.test_func_names.get(test_case_class, None)
        if test_fn_names is not None:
            # We already know them
            return test_fn_names

        test_fn_names = unittest.TestLoader.getTestCaseNames(self,
                                                             test_case_class)
        self.test_func_names[test_case_class] = test_fn_names
        return test_fn_names


class FilteredByModuleTestLoader(TestLoader):
    """A test loader that import only the needed modules."""

    def __init__(self, needs_module):
        """Constructor.

        :param needs_module: a callable taking a module name as a
            parameter returing True if the module should be loaded.
        """
        TestLoader.__init__(self)
        self.needs_module = needs_module

    def loadTestsFromModuleName(self, name):
        if self.needs_module(name):
            return TestLoader.loadTestsFromModuleName(self, name)
        else:
            return self.suiteClass()


class TestVisitor:
    """A visitor for Tests"""

    def visitSuite(self, aTestSuite):
        pass

    def visitCase(self, aTestCase):
        pass



# Copyright (C) 2008, 2010 Jelmer Vernooij <jelmer@samba.org>

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
"""Classify a commit based on the types of files it changed."""

import os.path

from ... import urlutils
from ...trace import mutter


def classify_filename(name):
    """Classify a file based on its name.

    :param name: File path.
    :return: One of code, documentation, translation or art.
        None if determining the file type failed.
    """
    # FIXME: Use mime types? Ohcount?
    # TODO: It will be better move those filters to properties file
    # and have possibility to determining own types !?
    extension = os.path.splitext(name)[1]
    if extension in (".c", ".h", ".py", ".cpp", ".rb", ".pm", ".pl", ".ac",
                     ".java", ".cc", ".proto", ".yy", ".l"):
        return "code"
    if extension in (".html", ".xml", ".txt", ".rst", ".TODO"):
        return "documentation"
    if extension in (".po",):
        return "translation"
    if extension in (".svg", ".png", ".jpg"):
        return "art"
    if not extension:
        basename = urlutils.basename(name)
        if basename in ("README", "NEWS", "TODO",
                        "AUTHORS", "COPYING"):
            return "documentation"
        if basename in ("Makefile",):
            return "code"

    mutter("don't know how to classify %s", name)
    return None


def classify_delta(delta):
    """Determine what sort of changes a delta contains.

    :param delta: A TreeDelta to inspect
    :return: List with classes found (see classify_filename)
    """
    # TODO: This is inaccurate, since it doesn't look at the
    # number of lines changed in a file.
    types = []
    for d in delta.added + delta.modified:
        types.append(classify_filename(d.path[1] or d.path[0]))
    return types



# Copyright (C) 2006-2011 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Test the lazy_import functionality."""

import linecache
import os
import re
import sys

from .. import errors, lazy_import, osutils
from . import TestCase, TestCaseInTempDir


class InstrumentedReplacer(lazy_import.ScopeReplacer):
    """Track what actions are done"""

    @staticmethod
    def use_actions(actions):
        InstrumentedReplacer.actions = actions

    def __getattribute__(self, attr):
        InstrumentedReplacer.actions.append(('__getattribute__', attr))
        return lazy_import.ScopeReplacer.__getattribute__(self, attr)

    def __call__(self, *args, **kwargs):
        InstrumentedReplacer.actions.append(('__call__', args, kwargs))
        return lazy_import.ScopeReplacer.__call__(self, *args, **kwargs)


class InstrumentedImportReplacer(lazy_import.ImportReplacer):

    @staticmethod
    def use_actions(actions):
        InstrumentedImportReplacer.actions = actions

    def _import(self, scope, name):
        InstrumentedImportReplacer.actions.append(('_import', name))
        return lazy_import.ImportReplacer._import(self, scope, name)

    def __getattribute__(self, attr):
        InstrumentedImportReplacer.actions.append(('__getattribute__', attr))
        return lazy_import.ScopeReplacer.__getattribute__(self, attr)

    def __call__(self, *args, **kwargs):
        InstrumentedImportReplacer.actions.append(('__call__', args, kwargs))
        return lazy_import.ScopeReplacer.__call__(self, *args, **kwargs)


class TestClass:
    """Just a simple test class instrumented for the test cases"""

    class_member = 'class_member'

    @staticmethod
    def use_actions(actions):
        TestClass.actions = actions

    def __init__(self):
        TestClass.actions.append('init')

    def foo(self, x):
        TestClass.actions.append(('foo', x))
        return 'foo'


class TestScopeReplacer(TestCase):
    """Test the ability of the replacer to put itself into the correct scope.

    In these tests we use the global scope, because we cannot replace
    variables in the local scope. This means that we need to be careful
    and not have the replacing objects use the same name, or we would
    get collisions.
    """

    def setUp(self):
        super().setUp()
        # These tests assume we will not be proxying, so make sure proxying is
        # disabled.
        orig_proxy = lazy_import.ScopeReplacer._should_proxy

        def restore():
            lazy_import.ScopeReplacer._should_proxy = orig_proxy
        lazy_import.ScopeReplacer._should_proxy = False

    def test_object(self):
        """ScopeReplacer can create an instance in local scope.

        An object should appear in globals() by constructing a ScopeReplacer,
        and it will be replaced with the real object upon the first request.
        """
        actions = []
        InstrumentedReplacer.use_actions(actions)
        TestClass.use_actions(actions)

        def factory(replacer, scope, name):
            actions.append('factory')
            return TestClass()

        try:
            test_obj1
        except NameError:
            # test_obj1 shouldn't exist yet
            pass
        else:
            self.fail('test_obj1 was not supposed to exist yet')

        InstrumentedReplacer(scope=globals(), name='test_obj1',
                             factory=factory)

        # We can't use isinstance() because that uses test_obj1.__class__
        # and that goes through __getattribute__ which would activate
        # the replacement
        self.assertEqual(InstrumentedReplacer,
                         object.__getattribute__(test_obj1, '__class__'))
        self.assertEqual('foo', test_obj1.foo(1))
        self.assertIsInstance(test_obj1, TestClass)
        self.assertEqual('foo', test_obj1.foo(2))
        self.assertEqual([('__getattribute__', 'foo'),
                          'factory',
                          'init',
                          ('foo', 1),
                          ('foo', 2),
                          ], actions)

    def test_setattr_replaces(self):
        """ScopeReplacer can create an instance in local scope.

        An object should appear in globals() by constructing a ScopeReplacer,
        and it will be replaced with the real object upon the first request.
        """
        actions = []
        TestClass.use_actions(actions)

        def factory(replacer, scope, name):
            return TestClass()
        try:
            test_obj6
        except NameError:
            # test_obj6 shouldn't exist yet
            pass
        else:
            self.fail('test_obj6 was not supposed to exist yet')

        lazy_import.ScopeReplacer(scope=globals(), name='test_obj6',
                                  factory=factory)

        # We can't use isinstance() because that uses test_obj6.__class__
        # and that goes through __getattribute__ which would activate
        # the replacement
        self.assertEqual(lazy_import.ScopeReplacer,
                         object.__getattribute__(test_obj6, '__class__'))
        test_obj6.bar = 'test'
        self.assertNotEqual(lazy_import.ScopeReplacer,
                            object.__getattribute__(test_obj6, '__class__'))
        self.assertEqual('test', test_obj6.bar)

    def test_replace_side_effects(self):
        """Creating a new object should only create one entry in globals.

        And only that entry even after replacement.
        """
        try:
            test_scope1
        except NameError:
            # test_scope1 shouldn't exist yet
            pass
        else:
            self.fail('test_scope1 was not supposed to exist yet')

        # ignore the logged actions
        TestClass.use_actions([])

        def factory(replacer, scope, name):
            return TestClass()

        orig_globals = set(globals().keys())

        lazy_import.ScopeReplacer(scope=globals(), name='test_scope1',
                                  factory=factory)

        new_globals = set(globals().keys())

        self.assertEqual(lazy_import.ScopeReplacer,
                         object.__getattribute__(test_scope1, '__class__'))
        self.assertEqual('foo', test_scope1.foo(1))
        self.assertIsInstance(test_scope1, TestClass)

        final_globals = set(globals().keys())

        self.assertEqual({'test_scope1'}, new_globals - orig_globals)
        self.assertEqual(set(), orig_globals - new_globals)
        self.assertEqual(set(), final_globals - new_globals)
        self.assertEqual(set(), new_globals - final_globals)

    def test_class(self):
        actions = []
        InstrumentedReplacer.use_actions(actions)
        TestClass.use_actions(actions)

        def factory(replacer, scope, name):
            actions.append('factory')
            return TestClass

        try:
            test_class1
        except NameError:
            # test_class2 shouldn't exist yet
            pass
        else:
            self.fail('test_class1 was not supposed to exist yet')

        InstrumentedReplacer(scope=globals(), name='test_class1',
                             factory=factory)

        self.assertEqual('class_member', test_class1.class_member)
        self.assertEqual(test_class1, TestClass)
        self.assertEqual([('__getattribute__', 'class_member'),
                          'factory',
                          ], actions)

    def test_call_class(self):
        actions = []
        InstrumentedReplacer.use_actions(actions)
        TestClass.use_actions(actions)

        def factory(replacer, scope, name):
            actions.append('factory')
            return TestClass

        try:
            test_class2
        except NameError:
            # test_class2 shouldn't exist yet
            pass
        else:
            self.fail('test_class2 was not supposed to exist yet')

        InstrumentedReplacer(scope=globals(), name='test_class2',
                             factory=factory)

        self.assertFalse(test_class2 is TestClass)
        obj = test_class2()
        self.assertIs(test_class2, TestClass)
        self.assertIsInstance(obj, TestClass)
        self.assertEqual('class_member', obj.class_member)
        self.assertEqual([('__call__', (), {}),
                          'factory',
                          'init',
                          ], actions)

    def test_call_func(self):
        actions = []
        InstrumentedReplacer.use_actions(actions)

        def func(a, b, c=None):
            actions.append('func')
            return (a, b, c)

        def factory(replacer, scope, name):
            actions.append('factory')
            return func

        try:
            test_func1
        except NameError:
            # test_func1 shouldn't exist yet
            pass
        else:
            self.fail('test_func1 was not supposed to exist yet')
        InstrumentedReplacer(scope=globals(), name='test_func1',
                             factory=factory)

        self.assertFalse(test_func1 is func)
        val = test_func1(1, 2, c='3')
        self.assertIs(test_func1, func)

        self.assertEqual((1, 2, '3'), val)
        self.assertEqual([('__call__', (1, 2), {'c': '3'}),
                          'factory',
                          'func',
                          ], actions)

    def test_other_variable(self):
        """Test when a ScopeReplacer is assigned to another variable.

        This test could be updated if we find a way to trap '=' rather
        than just giving a belated exception.
        ScopeReplacer only knows about the variable it was created as,
        so until the object is replaced, it is illegal to pass it to
        another variable. (Though discovering this may take a while)
        """
        actions = []
        InstrumentedReplacer.use_actions(actions)
        TestClass.use_actions(actions)

        def factory(replacer, scope, name):
            actions.append('factory')
            return TestClass()

        try:
            test_obj2
        except NameError:
            # test_obj2 shouldn't exist yet
            pass
        else:
            self.fail('test_obj2 was not supposed to exist yet')

        InstrumentedReplacer(scope=globals(), name='test_obj2',
                             factory=factory)

        self.assertEqual(InstrumentedReplacer,
                         object.__getattribute__(test_obj2, '__class__'))
        # This is technically not allowed, but we don't have a way to
        # test it until later.
        test_obj3 = test_obj2
        self.assertEqual(InstrumentedReplacer,
                         object.__getattribute__(test_obj2, '__class__'))
        self.assertEqual(InstrumentedReplacer,
                         object.__getattribute__(test_obj3, '__class__'))

        # The first use of the alternate variable causes test_obj2 to
        # be replaced.
        self.assertEqual('foo', test_obj3.foo(1))
        # test_obj2 has been replaced, but the ScopeReplacer has no
        # idea of test_obj3
        self.assertEqual(TestClass,
                         object.__getattribute__(test_obj2, '__class__'))
        self.assertEqual(InstrumentedReplacer,
                         object.__getattribute__(test_obj3, '__class__'))
        # We should be able to access test_obj2 attributes normally
        self.assertEqual('foo', test_obj2.foo(2))
        self.assertEqual('foo', test_obj2.foo(3))

        # However, the next access on test_obj3 should raise an error
        # because only now are we able to detect the problem.
        self.assertRaises(lazy_import.IllegalUseOfScopeReplacer,
                          getattr, test_obj3, 'foo')

        self.assertEqual([('__getattribute__', 'foo'),
                          'factory',
                          'init',
                          ('foo', 1),
                          ('foo', 2),
                          ('foo', 3),
                          ('__getattribute__', 'foo'),
                          ], actions)

    def test_enable_proxying(self):
        """Test that we can allow ScopeReplacer to proxy."""
        actions = []
        InstrumentedReplacer.use_actions(actions)
        TestClass.use_actions(actions)

        def factory(replacer, scope, name):
            actions.append('factory')
            return TestClass()

        try:
            test_obj4
        except NameError:
            # test_obj4 shouldn't exist yet
            pass
        else:
            self.fail('test_obj4 was not supposed to exist yet')

        lazy_import.ScopeReplacer._should_proxy = True
        InstrumentedReplacer(scope=globals(), name='test_obj4',
                             factory=factory)

        self.assertEqual(InstrumentedReplacer,
                         object.__getattribute__(test_obj4, '__class__'))
        test_obj5 = test_obj4
        self.assertEqual(InstrumentedReplacer,
                         object.__getattribute__(test_obj4, '__class__'))
        self.assertEqual(InstrumentedReplacer,
                         object.__getattribute__(test_obj5, '__class__'))

        # The first use of the alternate variable causes test_obj2 to
        # be replaced.
        self.assertEqual('foo', test_obj4.foo(1))
        self.assertEqual(TestClass,
                         object.__getattribute__(test_obj4, '__class__'))
        self.assertEqual(InstrumentedReplacer,
                         object.__getattribute__(test_obj5, '__class__'))
        # We should be able to access test_obj4 attributes normally
        self.assertEqual('foo', test_obj4.foo(2))
        # because we enabled proxying, test_obj5 can access its members as well
        self.assertEqual('foo', test_obj5.foo(3))
        self.assertEqual('foo', test_obj5.foo(4))

        # However, it cannot be replaced by the ScopeReplacer
        self.assertEqual(InstrumentedReplacer,
                         object.__getattribute__(test_obj5, '__class__'))

        self.assertEqual([('__getattribute__', 'foo'),
                          'factory',
                          'init',
                          ('foo', 1),
                          ('foo', 2),
                          ('__getattribute__', 'foo'),
                          ('foo', 3),
                          ('__getattribute__', 'foo'),
                          ('foo', 4),
                          ], actions)

    def test_replacing_from_own_scope_fails(self):
        """If a ScopeReplacer tries to replace itself a nice error is given"""
        actions = []
        InstrumentedReplacer.use_actions(actions)
        TestClass.use_actions(actions)

        def factory(replacer, scope, name):
            actions.append('factory')
            # return the name in given scope, which is currently the replacer
            return scope[name]

        try:
            test_obj7
        except NameError:
            # test_obj7 shouldn't exist yet
            pass
        else:
            self.fail('test_obj7 was not supposed to exist yet')

        InstrumentedReplacer(scope=globals(), name='test_obj7',
                             factory=factory)

        self.assertEqual(InstrumentedReplacer,
                         object.__getattribute__(test_obj7, '__class__'))
        e = self.assertRaises(lazy_import.IllegalUseOfScopeReplacer, test_obj7)
        self.assertIn("replace itself", e.msg)
        self.assertEqual([('__call__', (), {}),
                          'factory'], actions)


class ImportReplacerHelper(TestCaseInTempDir):
    """Test the ability to have a lazily imported module or object"""

    def setUp(self):
        super().setUp()
        self.create_modules()
        base_path = self.test_dir + '/base'

        self.actions = []
        InstrumentedImportReplacer.use_actions(self.actions)

        sys.path.append(base_path)
        self.addCleanup(sys.path.remove, base_path)

        def instrumented_import(mod, scope1, scope2, fromlist, level):
            self.actions.append(('import', mod, fromlist, level))
            return __import__(mod, scope1, scope2, fromlist, level=level)
        self.addCleanup(setattr, lazy_import, '_builtin_import', __import__)
        lazy_import._builtin_import = instrumented_import

    def create_modules(self):
        """Create some random modules to be imported.

        Each entry has a random suffix, and the full names are saved

        These are setup as follows:
         base/ <= used to ensure not in default search path
            root-XXX/
                __init__.py <= This will contain var1, func1
                mod-XXX.py <= This will contain var2, func2
                sub-XXX/
                    __init__.py <= Contains var3, func3
                    submoda-XXX.py <= contains var4, func4
                    submodb-XXX.py <= containse var5, func5
        """
        rand_suffix = osutils.rand_chars(4)
        root_name = 'root_' + rand_suffix
        mod_name = 'mod_' + rand_suffix
        sub_name = 'sub_' + rand_suffix
        submoda_name = 'submoda_' + rand_suffix
        submodb_name = 'submodb_' + rand_suffix

        os.mkdir('base')
        root_path = osutils.pathjoin('base', root_name)
        os.mkdir(root_path)
        root_init = osutils.pathjoin(root_path, '__init__.py')
        with open(osutils.pathjoin(root_path, '__init__.py'), 'w') as f:
            f.write('var1 = 1\ndef func1(a):\n  return a\n')
        mod_path = osutils.pathjoin(root_path, mod_name + '.py')
        with open(mod_path, 'w') as f:
            f.write('var2 = 2\ndef func2(a):\n  return a\n')

        sub_path = osutils.pathjoin(root_path, sub_name)
        os.mkdir(sub_path)
        with open(osutils.pathjoin(sub_path, '__init__.py'), 'w') as f:
            f.write('var3 = 3\ndef func3(a):\n  return a\n')
        submoda_path = osutils.pathjoin(sub_path, submoda_name + '.py')
        with open(submoda_path, 'w') as f:
            f.write('var4 = 4\ndef func4(a):\n  return a\n')
        submodb_path = osutils.pathjoin(sub_path, submodb_name + '.py')
        with open(submodb_path, 'w') as f:
            f.write('var5 = 5\ndef func5(a):\n  return a\n')
        self.root_name = root_name
        self.mod_name = mod_name
        self.sub_name = sub_name
        self.submoda_name = submoda_name
        self.submodb_name = submodb_name


class TestImportReplacerHelper(ImportReplacerHelper):

    def test_basic_import(self):
        """Test that a real import of these modules works"""
        sub_mod_path = '.'.join([self.root_name, self.sub_name,
                                 self.submoda_name])
        root = lazy_import._builtin_import(sub_mod_path, {}, {}, [], 0)
        self.assertEqual(1, root.var1)
        self.assertEqual(3, getattr(root, self.sub_name).var3)
        self.assertEqual(4, getattr(getattr(root, self.sub_name),
                                    self.submoda_name).var4)

        mod_path = '.'.join([self.root_name, self.mod_name])
        root = lazy_import._builtin_import(mod_path, {}, {}, [], 0)
        self.assertEqual(2, getattr(root, self.mod_name).var2)

        self.assertEqual([('import', sub_mod_path, [], 0),
                          ('import', mod_path, [], 0),
                          ], self.actions)


class TestImportReplacer(ImportReplacerHelper):

    def test_import_root(self):
        """Test 'import root-XXX as root1'"""
        try:
            root1
        except NameError:
            # root1 shouldn't exist yet
            pass
        else:
            self.fail('root1 was not supposed to exist yet')

        # This should replicate 'import root-xxyyzz as root1'
        InstrumentedImportReplacer(scope=globals(), name='root1',
                                   module_path=[self.root_name],
                                   member=None, children={})

        self.assertEqual(InstrumentedImportReplacer,
                         object.__getattribute__(root1, '__class__'))
        self.assertEqual(1, root1.var1)
        self.assertEqual('x', root1.func1('x'))

        self.assertEqual([('__getattribute__', 'var1'),
                          ('_import', 'root1'),
                          ('import', self.root_name, [], 0),
                          ], self.actions)

    def test_import_mod(self):
        """Test 'import root-XXX.mod-XXX as mod2'"""
        try:
            mod1
        except NameError:
            # mod1 shouldn't exist yet
            pass
        else:
            self.fail('mod1 was not supposed to exist yet')

        mod_path = self.root_name + '.' + self.mod_name
        InstrumentedImportReplacer(scope=globals(), name='mod1',
                                   module_path=[self.root_name, self.mod_name],
                                   member=None, children={})

        self.assertEqual(InstrumentedImportReplacer,
                         object.__getattribute__(mod1, '__class__'))
        self.assertEqual(2, mod1.var2)
        self.assertEqual('y', mod1.func2('y'))

        self.assertEqual([('__getattribute__', 'var2'),
                          ('_import', 'mod1'),
                          ('import', mod_path, [], 0),
                          ], self.actions)

    def test_import_mod_from_root(self):
        """Test 'from root-XXX import mod-XXX as mod2'"""
        try:
            mod2
        except NameError:
            # mod2 shouldn't exist yet
            pass
        else:
            self.fail('mod2 was not supposed to exist yet')

        InstrumentedImportReplacer(scope=globals(), name='mod2',
                                   module_path=[self.root_name],
                                   member=self.mod_name, children={})

        self.assertEqual(InstrumentedImportReplacer,
                         object.__getattribute__(mod2, '__class__'))
        self.assertEqual(2, mod2.var2)
        self.assertEqual('y', mod2.func2('y'))

        self.assertEqual([('__getattribute__', 'var2'),
                          ('_import', 'mod2'),
                          ('import', self.root_name, [self.mod_name], 0),
                          ], self.actions)

    def test_import_root_and_mod(self):
        """Test 'import root-XXX.mod-XXX' remapping both to root3.mod3"""
        try:
            root3
        except NameError:
            # root3 shouldn't exist yet
            pass
        else:
            self.fail('root3 was not supposed to exist yet')

        InstrumentedImportReplacer(scope=globals(),
                                   name='root3', module_path=[self.root_name], member=None,
                                   children={'mod3': ([self.root_name, self.mod_name], None, {})})

        # So 'root3' should be a lazy import
        # and once it is imported, mod3 should also be lazy until
        # actually accessed.
        self.assertEqual(InstrumentedImportReplacer,
                         object.__getattribute__(root3, '__class__'))
        self.assertEqual(1, root3.var1)

        # There is a mod3 member, but it is also lazy
        self.assertEqual(InstrumentedImportReplacer,
                         object.__getattribute__(root3.mod3, '__class__'))
        self.assertEqual(2, root3.mod3.var2)

        mod_path = self.root_name + '.' + self.mod_name
        self.assertEqual([('__getattribute__', 'var1'),
                          ('_import', 'root3'),
                          ('import', self.root_name, [], 0),
                          ('__getattribute__', 'var2'),
                          ('_import', 'mod3'),
                          ('import', mod_path, [], 0),
                          ], self.actions)

    def test_import_root_and_root_mod(self):
        """Test that 'import root, root.mod' can be done.

        The second import should re-use the first one, and just add
        children to be imported.
        """
        try:
            root4
        except NameError:
            # root4 shouldn't exist yet
            pass
        else:
            self.fail('root4 was not supposed to exist yet')

        InstrumentedImportReplacer(scope=globals(),
                                   name='root4', module_path=[self.root_name], member=None,
                                   children={})

        # So 'root4' should be a lazy import
        self.assertEqual(InstrumentedImportReplacer,
                         object.__getattribute__(root4, '__class__'))

        # Lets add a new child to be imported on demand
        # This syntax of using object.__getattribute__ is the correct method
        # for accessing the _import_replacer_children member
        children = object.__getattribute__(root4, '_import_replacer_children')
        children['mod4'] = ([self.root_name, self.mod_name], None, {})

        # Accessing root4.mod4 should import root, but mod should stay lazy
        self.assertEqual(InstrumentedImportReplacer,
                         object.__getattribute__(root4.mod4, '__class__'))
        self.assertEqual(2, root4.mod4.var2)

        mod_path = self.root_name + '.' + self.mod_name
        self.assertEqual([('__getattribute__', 'mod4'),
                          ('_import', 'root4'),
                          ('import', self.root_name, [], 0),
                          ('__getattribute__', 'var2'),
                          ('_import', 'mod4'),
                          ('import', mod_path, [], 0),
                          ], self.actions)

    def test_import_root_sub_submod(self):
        """Test import root.mod, root.sub.submoda, root.sub.submodb
        root should be a lazy import, with multiple children, who also
        have children to be imported.
        And when root is imported, the children should be lazy, and
        reuse the intermediate lazy object.
        """
        try:
            root5
        except NameError:
            # root5 shouldn't exist yet
            pass
        else:
            self.fail('root5 was not supposed to exist yet')

        InstrumentedImportReplacer(scope=globals(),
                                   name='root5', module_path=[self.root_name], member=None,
                                   children={'mod5': ([self.root_name, self.mod_name], None, {}),
                                             'sub5': ([self.root_name, self.sub_name], None,
                                                      {'submoda5': ([self.root_name, self.sub_name,
                                                                     self.submoda_name], None, {}),
                                                       'submodb5': ([self.root_name, self.sub_name,
                                                                     self.submodb_name], None, {})
                                                       }),
                                             })

        # So 'root5' should be a lazy import
        self.assertEqual(InstrumentedImportReplacer,
                         object.__getattribute__(root5, '__class__'))

        # Accessing root5.mod5 should import root, but mod should stay lazy
        self.assertEqual(InstrumentedImportReplacer,
                         object.__getattribute__(root5.mod5, '__class__'))
        # root5.sub5 should still be lazy, but not re-import root5
        self.assertEqual(InstrumentedImportReplacer,
                         object.__getattribute__(root5.sub5, '__class__'))

        # Accessing root5.sub5.submoda5 should import sub5, but not either
        # of the sub objects (they should be available as lazy objects
        self.assertEqual(InstrumentedImportReplacer,
                         object.__getattribute__(root5.sub5.submoda5, '__class__'))
        self.assertEqual(InstrumentedImportReplacer,
                         object.__getattribute__(root5.sub5.submodb5, '__class__'))

        # This should import mod5
        self.assertEqual(2, root5.mod5.var2)
        # These should import submoda5 and submodb5
        self.assertEqual(4, root5.sub5.submoda5.var4)
        self.assertEqual(5, root5.sub5.submodb5.var5)

        mod_path = self.root_name + '.' + self.mod_name
        sub_path = self.root_name + '.' + self.sub_name
        submoda_path = sub_path + '.' + self.submoda_name
        submodb_path = sub_path + '.' + self.submodb_name

        self.assertEqual([('__getattribute__', 'mod5'),
                          ('_import', 'root5'),
                          ('import', self.root_name, [], 0),
                          ('__getattribute__', 'submoda5'),
                          ('_import', 'sub5'),
                          ('import', sub_path, [], 0),
                          ('__getattribute__', 'var2'),
                          ('_import', 'mod5'),
                          ('import', mod_path, [], 0),
                          ('__getattribute__', 'var4'),
                          ('_import', 'submoda5'),
                          ('import', submoda_path, [], 0),
                          ('__getattribute__', 'var5'),
                          ('_import', 'submodb5'),
                          ('import', submodb_path, [], 0),
                          ], self.actions)


class TestConvertImportToMap(TestCase):
    """Directly test the conversion from import strings to maps"""

    def check(self, expected, import_strings):
        proc = lazy_import.ImportProcessor()
        for import_str in import_strings:
            proc._convert_import_str(import_str)
        self.assertEqual(expected, proc.imports,
                         'Import of %r was not converted correctly'
                         ' %s != %s' % (import_strings, expected,
                                        proc.imports))

    def test_import_one(self):
        self.check({'one': (['one'], None, {}),
                    }, ['import one'])

    def test_import_one_two(self):
        one_two_map = {'one': (['one'], None,
                               {'two': (['one', 'two'], None, {}),
                                }),
                       }
        self.check(one_two_map, ['import one.two'])
        self.check(one_two_map, ['import one, one.two'])
        self.check(one_two_map, ['import one', 'import one.two'])
        self.check(one_two_map, ['import one.two', 'import one'])

    def test_import_one_two_three(self):
        one_two_three_map = {
            'one': (['one'], None,
                    {'two': (['one', 'two'], None,
                             {'three': (['one', 'two', 'three'], None, {}),
                              }),
                     }),
        }
        self.check(one_two_three_map, ['import one.two.three'])
        self.check(one_two_three_map, ['import one, one.two.three'])
        self.check(one_two_three_map, ['import one',
                                       'import one.two.three'])
        self.check(one_two_three_map, ['import one.two.three',
                                       'import one'])

    def test_import_one_as_x(self):
        self.check({'x': (['one'], None, {}),
                    }, ['import one as x'])

    def test_import_one_two_as_x(self):
        self.check({'x': (['one', 'two'], None, {}),
                    }, ['import one.two as x'])

    def test_import_mixed(self):
        mixed = {'x': (['one', 'two'], None, {}),
                 'one': (['one'], None,
                         {'two': (['one', 'two'], None, {}),
                          }),
                 }
        self.check(mixed, ['import one.two as x, one.two'])
        self.check(mixed, ['import one.two as x', 'import one.two'])
        self.check(mixed, ['import one.two', 'import one.two as x'])

    def test_import_with_as(self):
        self.check({'fast': (['fast'], None, {})}, ['import fast'])


class TestFromToMap(TestCase):
    """Directly test the conversion of 'from foo import bar' syntax"""

    def check_result(self, expected, from_strings):
        proc = lazy_import.ImportProcessor()
        for from_str in from_strings:
            proc._convert_from_str(from_str)
        self.assertEqual(expected, proc.imports,
                         'Import of %r was not converted correctly'
                         ' %s != %s' % (from_strings, expected, proc.imports))

    def test_from_one_import_two(self):
        self.check_result({'two': (['one'], 'two', {})},
                          ['from one import two'])

    def test_from_one_import_two_as_three(self):
        self.check_result({'three': (['one'], 'two', {})},
                          ['from one import two as three'])

    def test_from_one_import_two_three(self):
        two_three_map = {'two': (['one'], 'two', {}),
                         'three': (['one'], 'three', {}),
                         }
        self.check_result(two_three_map,
                          ['from one import two, three'])
        self.check_result(two_three_map,
                          ['from one import two',
                           'from one import three'])

    def test_from_one_two_import_three(self):
        self.check_result({'three': (['one', 'two'], 'three', {})},
                          ['from one.two import three'])


class TestCanonicalize(TestCase):
    """Test that we can canonicalize import texts"""

    def check(self, expected, text):
        proc = lazy_import.ImportProcessor()
        parsed = proc._canonicalize_import_text(text)
        self.assertEqual(expected, parsed,
                         'Incorrect parsing of text:\n%s\n%s\n!=\n%s'
                         % (text, expected, parsed))

    def test_import_one(self):
        self.check(['import one'], 'import one')
        self.check(['import one'], '\nimport one\n\n')

    def test_import_one_two(self):
        self.check(['import one, two'], 'import one, two')
        self.check(['import one, two'], '\nimport one, two\n\n')

    def test_import_one_as_two_as(self):
        self.check(['import one as x, two as y'], 'import one as x, two as y')
        self.check(['import one as x, two as y'],
                   '\nimport one as x, two as y\n')

    def test_from_one_import_two(self):
        self.check(['from one import two'], 'from one import two')
        self.check(['from one import two'], '\nfrom one import two\n\n')
        self.check(['from one import two'], '\nfrom one import (two)\n')
        self.check(['from one import  two '],
                   '\nfrom one import (\n\ttwo\n)\n')

    def test_multiple(self):
        self.check(['import one', 'import two, three', 'from one import four'],
                   'import one\nimport two, three\nfrom one import four')
        self.check(['import one', 'import two, three', 'from one import four'],
                   'import one\nimport (two, three)\nfrom one import four')
        self.check(['import one', 'import two, three', 'from one import four'],
                   'import one\n'
                   'import two, three\n'
                   'from one import four')
        self.check(['import one',
                    'import two, three', 'from one import  four, '],
                   'import one\n'
                   'import two, three\n'
                   'from one import (\n'
                   '    four,\n'
                   '    )\n'
                   )

    def test_missing_trailing(self):
        proc = lazy_import.ImportProcessor()
        self.assertRaises(lazy_import.InvalidImportLine,
                          proc._canonicalize_import_text,
                          "from foo import (\n  bar\n")


class TestImportProcessor(TestCase):
    """Test that ImportProcessor can turn import texts into lazy imports"""

    def check(self, expected, text):
        proc = lazy_import.ImportProcessor()
        proc._build_map(text)
        self.assertEqual(expected, proc.imports,
                         'Incorrect processing of:\n%s\n%s\n!=\n%s'
                         % (text, expected, proc.imports))

    def test_import_one(self):
        exp = {'one': (['one'], None, {})}
        self.check(exp, 'import one')
        self.check(exp, '\nimport one\n')

    def test_import_one_two(self):
        exp = {'one': (['one'], None,
                       {'two': (['one', 'two'], None, {}),
                        }),
               }
        self.check(exp, 'import one.two')
        self.check(exp, 'import one, one.two')
        self.check(exp, 'import one\nimport one.two')

    def test_import_as(self):
        exp = {'two': (['one'], None, {})}
        self.check(exp, 'import one as two')

    def test_import_many(self):
        exp = {'one': (['one'], None,
                       {'two': (['one', 'two'], None,
                                {'three': (['one', 'two', 'three'], None, {}),
                                 }),
                        'four': (['one', 'four'], None, {}),
                        }),
               'five': (['one', 'five'], None, {}),
               }
        self.check(exp, 'import one.two.three, one.four, one.five as five')
        self.check(exp, 'import one.five as five\n'
                        'import one\n'
                        'import one.two.three\n'
                        'import one.four\n')

    def test_from_one_import_two(self):
        exp = {'two': (['one'], 'two', {})}
        self.check(exp, 'from one import two\n')
        self.check(exp, 'from one import (\n'
                        '    two,\n'
                        '    )\n')

    def test_from_one_import_two_two(self):
        exp = {'two': (['one'], 'two', {})}
        self.check(exp, 'from one import two\n')
        self.check(exp, 'from one import (two)\n')
        self.check(exp, 'from one import (two,)\n')
        self.check(exp, 'from one import two as two\n')
        self.check(exp, 'from one import (\n'
                        '    two,\n'
                        '    )\n')

    def test_from_many(self):
        exp = {'two': (['one'], 'two', {}),
               'three': (['one', 'two'], 'three', {}),
               'five': (['one', 'two'], 'four', {}),
               }
        self.check(exp, 'from one import two\n'
                        'from one.two import three, four as five\n')
        self.check(exp, 'from one import two\n'
                        'from one.two import (\n'
                        '    three,\n'
                        '    four as five,\n'
                        '    )\n')

    def test_mixed(self):
        exp = {'two': (['one'], 'two', {}),
               'three': (['one', 'two'], 'three', {}),
               'five': (['one', 'two'], 'four', {}),
               'one': (['one'], None,
                       {'two': (['one', 'two'], None, {}),
                        }),
               }
        self.check(exp, 'from one import two\n'
                        'from one.two import three, four as five\n'
                        'import one.two')
        self.check(exp, 'from one import two\n'
                        'from one.two import (\n'
                        '    three,\n'
                        '    four as five,\n'
                        '    )\n'
                        'import one\n'
                        'import one.two\n')

    def test_incorrect_line(self):
        proc = lazy_import.ImportProcessor()
        self.assertRaises(lazy_import.InvalidImportLine,
                          proc._build_map, 'foo bar baz')
        self.assertRaises(lazy_import.InvalidImportLine,
                          proc._build_map, 'improt foo')
        self.assertRaises(lazy_import.InvalidImportLine,
                          proc._build_map, 'importfoo')
        self.assertRaises(lazy_import.InvalidImportLine,
                          proc._build_map, 'fromimport')

    def test_name_collision(self):
        proc = lazy_import.ImportProcessor()
        proc._build_map('import foo')

        # All of these would try to create an object with the
        # same name as an existing object.
        self.assertRaises(lazy_import.ImportNameCollision,
                          proc._build_map, 'import bar as foo')
        self.assertRaises(lazy_import.ImportNameCollision,
                          proc._build_map, 'from foo import bar as foo')
        self.assertRaises(lazy_import.ImportNameCollision,
                          proc._build_map, 'from bar import foo')

    def test_relative_imports(self):
        proc = lazy_import.ImportProcessor()
        self.assertRaises(ImportError,
                          proc._build_map, 'import .bar as foo')
        self.assertRaises(ImportError,
                          proc._build_map, 'from .foo import bar as foo')
        self.assertRaises(ImportError,
                          proc._build_map, 'from .bar import foo')


class TestLazyImportProcessor(ImportReplacerHelper):

    def test_root(self):
        try:
            root6
        except NameError:
            pass  # root6 should not be defined yet
        else:
            self.fail('root6 was not supposed to exist yet')

        text = 'import {} as root6'.format(self.root_name)
        proc = lazy_import.ImportProcessor(InstrumentedImportReplacer)
        proc.lazy_import(scope=globals(), text=text)

        # So 'root6' should be a lazy import
        self.assertEqual(InstrumentedImportReplacer,
                         object.__getattribute__(root6, '__class__'))

        self.assertEqual(1, root6.var1)
        self.assertEqual('x', root6.func1('x'))

        self.assertEqual([('__getattribute__', 'var1'),
                          ('_import', 'root6'),
                          ('import', self.root_name, [], 0),
                          ], self.actions)

    def test_import_deep(self):
        """Test import root.mod, root.sub.submoda, root.sub.submodb
        root should be a lazy import, with multiple children, who also
        have children to be imported.
        And when root is imported, the children should be lazy, and
        reuse the intermediate lazy object.
        """
        try:
            submoda7
        except NameError:
            pass  # submoda7 should not be defined yet
        else:
            self.fail('submoda7 was not supposed to exist yet')

        text = """\
import %(root_name)s.%(sub_name)s.%(submoda_name)s as submoda7
""" % self.__dict__
        proc = lazy_import.ImportProcessor(InstrumentedImportReplacer)
        proc.lazy_import(scope=globals(), text=text)

        # So 'submoda7' should be a lazy import
        self.assertEqual(InstrumentedImportReplacer,
                         object.__getattribute__(submoda7, '__class__'))

        # This should import submoda7
        self.assertEqual(4, submoda7.var4)

        sub_path = self.root_name + '.' + self.sub_name
        submoda_path = sub_path + '.' + self.submoda_name

        self.assertEqual([('__getattribute__', 'var4'),
                          ('_import', 'submoda7'),
                          ('import', submoda_path, [], 0),
                          ], self.actions)

    def test_lazy_import(self):
        """Smoke test that lazy_import() does the right thing"""
        try:
            root8
        except NameError:
            pass  # root8 should not be defined yet
        else:
            self.fail('root8 was not supposed to exist yet')
        lazy_import.lazy_import(globals(),
                                'import {} as root8'.format(self.root_name),
                                lazy_import_class=InstrumentedImportReplacer)

        self.assertEqual(InstrumentedImportReplacer,
                         object.__getattribute__(root8, '__class__'))

        self.assertEqual(1, root8.var1)
        self.assertEqual(1, root8.var1)
        self.assertEqual(1, root8.func1(1))

        self.assertEqual([('__getattribute__', 'var1'),
                          ('_import', 'root8'),
                          ('import', self.root_name, [], 0),
                          ], self.actions)


class TestScopeReplacerReentrance(TestCase):
    """The ScopeReplacer should be reentrant.

    Invoking a replacer while an invocation was already on-going leads to a
    race to see which invocation will be the first to call _replace.
    The losing caller used to see an exception (bugs 396819 and 702914).

    These tests set up a tracer that stops at a suitable moment (upon
    entry of a specified method) and starts another call to the
    functionality in question (__call__, __getattribute__, __setattr_)
    in order to win the race, setting up the original caller to lose.
    """

    def tracer(self, frame, event, arg):
        if event != 'call':
            return self.tracer
        # Grab the name of the file that contains the code being executed.
        code = frame.f_code
        filename = code.co_filename
        # Convert ".pyc" and ".pyo" file names to their ".py" equivalent.
        filename = re.sub(r'\.py[co]$', '.py', filename)
        function_name = code.co_name
        # If we're executing a line of code from the right module...
        if (filename.endswith('lazy_import.py')
                and function_name == self.method_to_trace):
            # We don't need to trace any more.
            sys.settrace(None)
            # Run another racer.  This one will "win" the race.
            self.racer()
        return self.tracer

    def run_race(self, racer, method_to_trace='_resolve'):
        self.overrideAttr(lazy_import.ScopeReplacer, '_should_proxy', True)
        self.racer = racer
        self.method_to_trace = method_to_trace
        sys.settrace(self.tracer)
        self.racer()  # Should not raise any exception
        # Make sure the tracer actually found the code it was
        # looking for.  If not, maybe the code was refactored in
        # such a way that these tests aren't needed any more.
        self.assertEqual(None, sys.gettrace())

    def test_call(self):
        def factory(*args):
            return factory
        replacer = lazy_import.ScopeReplacer({}, factory, 'name')
        self.run_race(replacer)

    def test_setattr(self):
        class Replaced:
            pass

        def factory(*args):
            return Replaced()

        replacer = lazy_import.ScopeReplacer({}, factory, 'name')

        def racer():
            replacer.foo = 42

        self.run_race(racer)

    def test_getattribute(self):
        class Replaced:
            foo = 'bar'

        def factory(*args):
            return Replaced()

        replacer = lazy_import.ScopeReplacer({}, factory, 'name')

        def racer():
            replacer.foo

        self.run_race(racer)



# Copyright (C) 2006, 2007, 2009, 2011 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

from .. import branch, errors
from . import TestCaseWithTransport


class TestExtract(TestCaseWithTransport):

    def test_extract(self):
        self.build_tree(['a/', 'a/b/', 'a/b/c', 'a/d'])
        wt = self.make_branch_and_tree('a', format='rich-root-pack')
        wt.add(['b', 'b/c', 'd'], ids=[b'b-id', b'c-id', b'd-id'])
        wt.commit('added files')
        b_wt = wt.extract('b')
        self.assertTrue(b_wt.is_versioned(''))
        if b_wt.supports_setting_file_ids():
            self.assertEqual(b'b-id', b_wt.path2id(''))
            self.assertEqual(b'c-id', b_wt.path2id('c'))
            self.assertEqual('c', b_wt.id2path(b'c-id'))
            self.assertRaises(errors.BzrError, wt.id2path, b'b-id')
        self.assertEqual(b_wt.basedir, wt.abspath('b'))
        self.assertEqual(wt.get_parent_ids(), b_wt.get_parent_ids())
        self.assertEqual(wt.branch.last_revision(),
                         b_wt.branch.last_revision())

    def extract_in_checkout(self, a_branch):
        self.build_tree(['a/', 'a/b/', 'a/b/c/', 'a/b/c/d'])
        wt = a_branch.create_checkout('a', lightweight=True)
        wt.add(['b', 'b/c', 'b/c/d'], ids=[b'b-id', b'c-id', b'd-id'])
        wt.commit('added files')
        return wt.extract('b')

    def test_extract_in_checkout(self):
        a_branch = self.make_branch('branch', format='rich-root-pack')
        self.extract_in_checkout(a_branch)
        b_branch = branch.Branch.open('branch/b')
        b_branch_ref = branch.Branch.open('a/b')
        self.assertEqual(b_branch.base, b_branch_ref.base)

    def test_extract_in_deep_checkout(self):
        a_branch = self.make_branch('branch', format='rich-root-pack')
        self.build_tree(['a/', 'a/b/', 'a/b/c/', 'a/b/c/d/', 'a/b/c/d/e'])
        wt = a_branch.create_checkout('a', lightweight=True)
        wt.add(['b', 'b/c', 'b/c/d', 'b/c/d/e/'], ids=[b'b-id', b'c-id', b'd-id', b'e-id'])
        wt.commit('added files')
        b_wt = wt.extract('b/c/d')
        b_branch = branch.Branch.open('branch/b/c/d')
        b_branch_ref = branch.Branch.open('a/b/c/d')
        self.assertEqual(b_branch.base, b_branch_ref.base)

    def test_bad_repo_format(self):
        repo = self.make_repository('branch', shared=True,
                                    format='knit')
        a_branch = repo.controldir.create_branch()
        self.assertRaises(errors.RootNotRich, self.extract_in_checkout,
                          a_branch)

    def test_good_repo_format(self):
        repo = self.make_repository('branch', shared=True,
                                    format='dirstate-with-subtree')
        a_branch = repo.controldir.create_branch()
        wt_b = self.extract_in_checkout(a_branch)
        self.assertEqual(wt_b.branch.repository.controldir.transport.base,
                         repo.controldir.transport.base)



# Copyright (C) 2005, 2006 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""API Documentation for breezy.

This documentation is made up of doctest testable examples.

Look for `breezy/doc/api/*.txt` to read it.

This documentation documents the current best practice in using the library.
For details on specific apis, see pydoc on the api, or read the source.
"""

import doctest
import os

from breezy import tests


def make_new_test_id(test):
    new_id = '{}.DocFileTest({})'.format(__name__, test.id())
    return lambda: new_id


def load_tests(loader, basic_tests, pattern):
    """This module creates its own test suite with DocFileSuite."""

    dir_ = os.path.dirname(__file__)
    if os.path.isdir(dir_):
        candidates = os.listdir(dir_)
    else:
        candidates = []
    scripts = [candidate for candidate in candidates
               if candidate.endswith('.txt')]
    # since this module doesn't define tests, we ignore basic_tests
    suite = doctest.DocFileSuite(*scripts, setUp=tests.isolated_doctest_setUp,
                                 tearDown=tests.isolated_doctest_tearDown)
    # DocFileCase reduces the test id to the base name of the tested file, we
    # want the module to appears there.
    for t in tests.iter_suite_tests(suite):
        t.id = make_new_test_id(t)
    return suite



# Copyright (C) 2007, 2009-2012, 2016 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Tests for Branch.get_rev_id."""

from breezy.errors import RevnoOutOfBounds
from breezy.revision import NULL_REVISION
from breezy.tests import TestCaseWithTransport


class TestGetRevid(TestCaseWithTransport):

    def test_empty_branch(self):
        # on an empty branch we want (0, NULL_REVISION)
        branch = self.make_branch('branch')
        self.assertEqual(NULL_REVISION, branch.get_rev_id(0))
        self.assertRaises(RevnoOutOfBounds, branch.get_rev_id, 1)
        self.assertRaises(RevnoOutOfBounds, branch.get_rev_id, -1)

    def test_non_empty_branch(self):
        # after the second commit we want (2, 'second-revid')
        tree = self.make_branch_and_tree('branch')
        revid1 = tree.commit('1st post')
        revid2 = tree.commit('2st post', allow_pointless=True)
        self.assertEqual(revid2, tree.branch.get_rev_id(2))
        self.assertEqual(revid1, tree.branch.get_rev_id(1))
        self.assertRaises(RevnoOutOfBounds, tree.branch.get_rev_id, 3)



# Copyright (C) 2006-2011 Robey Pointer <robey@lag.net>
# Copyright (C) 2005, 2006, 2007 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Foundation SSH support for SFTP and smart server."""

import errno
import getpass
import logging
import os
import socket
import subprocess
import sys
from binascii import hexlify
from typing import Dict, Optional, Set, Tuple, Type

from .. import bedding, config, errors, osutils, trace, ui

try:
    import paramiko
except ModuleNotFoundError as e:
    # If we have an ssh subprocess, we don't strictly need paramiko for all ssh
    # access
    paramiko = None  # type: ignore
else:
    from paramiko.sftp_client import SFTPClient


class StrangeHostname(errors.BzrError):
    _fmt = "Refusing to connect to strange SSH hostname %(hostname)s"


SYSTEM_HOSTKEYS: Dict[str, Dict[str, str]] = {}
BRZ_HOSTKEYS: Dict[str, Dict[str, str]] = {}


class SSHVendorManager:
    """Manager for manage SSH vendors."""

    # Note, although at first sign the class interface seems similar to
    # breezy.registry.Registry it is not possible/convenient to directly use
    # the Registry because the class just has "get()" interface instead of the
    # Registry's "get(key)".

    def __init__(self):
        self._ssh_vendors = {}
        self._cached_ssh_vendor = None
        self._default_ssh_vendor = None

    def register_default_vendor(self, vendor):
        """Register default SSH vendor."""
        self._default_ssh_vendor = vendor

    def register_vendor(self, name, vendor):
        """Register new SSH vendor by name."""
        self._ssh_vendors[name] = vendor

    def clear_cache(self):
        """Clear previously cached lookup result."""
        self._cached_ssh_vendor = None

    def _get_vendor_by_config(self):
        vendor_name = config.GlobalStack().get('ssh')
        if vendor_name is not None:
            try:
                vendor = self._ssh_vendors[vendor_name]
            except KeyError:
                vendor = self._get_vendor_from_path(vendor_name)
                if vendor is None:
                    raise errors.UnknownSSH(vendor_name)
                vendor.executable_path = vendor_name
            return vendor
        return None

    def _get_ssh_version_string(self, args):
        """Return SSH version string from the subprocess."""
        try:
            p = subprocess.Popen(args,
                                 stdout=subprocess.PIPE,
                                 stderr=subprocess.PIPE,
                                 bufsize=0,
                                 **os_specific_subprocess_params())
            stdout, stderr = p.communicate()
        except OSError:
            stdout = stderr = b''
        return (stdout + stderr).decode(osutils.get_terminal_encoding())

    def _get_vendor_by_version_string(self, version, progname):
        """Return the vendor or None based on output from the subprocess.

        :param version: The output of 'ssh -V' like command.
        :param args: Command line that was run.
        """
        vendor = None
        if 'OpenSSH' in version:
            trace.mutter('ssh implementation is OpenSSH')
            vendor = OpenSSHSubprocessVendor()
        elif 'SSH Secure Shell' in version:
            trace.mutter('ssh implementation is SSH Corp.')
            vendor = SSHCorpSubprocessVendor()
        elif 'lsh' in version:
            trace.mutter('ssh implementation is GNU lsh.')
            vendor = LSHSubprocessVendor()
        # As plink user prompts are not handled currently, don't auto-detect
        # it by inspection below, but keep this vendor detection for if a path
        # is given in BRZ_SSH. See https://bugs.launchpad.net/bugs/414743
        elif 'plink' in version and progname == 'plink':
            # Checking if "plink" was the executed argument as Windows
            # sometimes reports 'ssh -V' incorrectly with 'plink' in its
            # version.  See https://bugs.launchpad.net/bzr/+bug/107155
            trace.mutter("ssh implementation is Putty's plink.")
            vendor = PLinkSubprocessVendor()
        return vendor

    def _get_vendor_by_inspection(self):
        """Return the vendor or None by checking for known SSH implementations."""
        version = self._get_ssh_version_string(['ssh', '-V'])
        return self._get_vendor_by_version_string(version, "ssh")

    def _get_vendor_from_path(self, path):
        """Return the vendor or None using the program at the given path"""
        version = self._get_ssh_version_string([path, '-V'])
        return self._get_vendor_by_version_string(version,
                                                  os.path.splitext(os.path.basename(path))[0])

    def get_vendor(self):
        """Find out what version of SSH is on the system.

        :raises SSHVendorNotFound: if no any SSH vendor is found
        :raises UnknownSSH: if the BRZ_SSH environment variable contains
                            unknown vendor name
        """
        if self._cached_ssh_vendor is None:
            vendor = self._get_vendor_by_config()
            if vendor is None:
                vendor = self._get_vendor_by_inspection()
                if vendor is None:
                    trace.mutter('falling back to default implementation')
                    vendor = self._default_ssh_vendor
                    if vendor is None:
                        raise errors.SSHVendorNotFound()
            self._cached_ssh_vendor = vendor
        return self._cached_ssh_vendor


_ssh_vendor_manager = SSHVendorManager()
_get_ssh_vendor = _ssh_vendor_manager.get_vendor
register_default_ssh_vendor = _ssh_vendor_manager.register_default_vendor
register_ssh_vendor = _ssh_vendor_manager.register_vendor


def _ignore_signals():
    # TODO: This should possibly ignore SIGHUP as well, but bzr currently
    # doesn't handle it itself.
    # <https://launchpad.net/products/bzr/+bug/41433/+index>
    import signal
    signal.signal(signal.SIGINT, signal.SIG_IGN)
    # GZ 2010-02-19: Perhaps make this check if breakin is installed instead
    if signal.getsignal(signal.SIGQUIT) != signal.SIG_DFL:
        signal.signal(signal.SIGQUIT, signal.SIG_IGN)


class SocketAsChannelAdapter:
    """Simple wrapper for a socket that pretends to be a paramiko Channel."""

    def __init__(self, sock):
        self.__socket = sock

    def get_name(self):
        return "bzr SocketAsChannelAdapter"

    def send(self, data):
        return self.__socket.send(data)

    def recv(self, n):
        try:
            return self.__socket.recv(n)
        except OSError as e:
            if e.args[0] in (errno.EPIPE, errno.ECONNRESET, errno.ECONNABORTED,
                             errno.EBADF):
                # Connection has closed.  Paramiko expects an empty string in
                # this case, not an exception.
                return ''
            raise

    def recv_ready(self):
        # TODO: jam 20051215 this function is necessary to support the
        # pipelined() function. In reality, it probably should use
        # poll() or select() to actually return if there is data
        # available, otherwise we probably don't get any benefit
        return True

    def close(self):
        self.__socket.close()


class SSHVendor:
    """Abstract base class for SSH vendor implementations."""

    def connect_sftp(self, username, password, host, port):
        """Make an SSH connection, and return an SFTPClient.

        :param username: an ascii string
        :param password: an ascii string
        :param host: a host name as an ascii string
        :param port: a port number
        :type port: int

        :raises: ConnectionError if it cannot connect.

        :rtype: paramiko.sftp_client.SFTPClient
        """
        raise NotImplementedError(self.connect_sftp)

    def connect_ssh(self, username, password, host, port, command):
        """Make an SSH connection.

        :returns: an SSHConnection.
        """
        raise NotImplementedError(self.connect_ssh)

    def _raise_connection_error(self, host, port=None, orig_error=None,
                                msg='Unable to connect to SSH host'):
        """Raise a SocketConnectionError with properly formatted host.

        This just unifies all the locations that try to raise ConnectionError,
        so that they format things properly.
        """
        raise errors.SocketConnectionError(host=host, port=port, msg=msg,
                                           orig_error=orig_error)


class LoopbackVendor(SSHVendor):
    """SSH "vendor" that connects over a plain TCP socket, not SSH."""

    def connect_sftp(self, username, password, host, port):
        sock = socket.socket()
        try:
            sock.connect((host, port))
        except OSError as e:
            self._raise_connection_error(host, port=port, orig_error=e)
        return SFTPClient(SocketAsChannelAdapter(sock))


register_ssh_vendor('loopback', LoopbackVendor())


class ParamikoVendor(SSHVendor):
    """Vendor that uses paramiko."""

    def _hexify(self, s):
        return hexlify(s).upper()

    def _connect(self, username, password, host, port):
        global SYSTEM_HOSTKEYS, BRZ_HOSTKEYS

        load_host_keys()

        try:
            t = paramiko.Transport((host, port or 22))
            t.set_log_channel('bzr.paramiko')
            t.start_client()
        except (paramiko.SSHException, OSError) as e:
            self._raise_connection_error(host, port=port, orig_error=e)

        server_key = t.get_remote_server_key()
        server_key_hex = self._hexify(server_key.get_fingerprint())
        keytype = server_key.get_name()
        if host in SYSTEM_HOSTKEYS and keytype in SYSTEM_HOSTKEYS[host]:
            our_server_key = SYSTEM_HOSTKEYS[host][keytype]
            our_server_key_hex = self._hexify(our_server_key.get_fingerprint())
        elif host in BRZ_HOSTKEYS and keytype in BRZ_HOSTKEYS[host]:
            our_server_key = BRZ_HOSTKEYS[host][keytype]
            our_server_key_hex = self._hexify(our_server_key.get_fingerprint())
        else:
            trace.warning('Adding %s host key for %s: %s'
                          % (keytype, host, server_key_hex))
            add = getattr(BRZ_HOSTKEYS, 'add', None)
            if add is not None:  # paramiko >= 1.X.X
                BRZ_HOSTKEYS.add(host, keytype, server_key)
            else:
                BRZ_HOSTKEYS.setdefault(host, {})[keytype] = server_key
            our_server_key = server_key
            our_server_key_hex = self._hexify(our_server_key.get_fingerprint())
            save_host_keys()
        if server_key != our_server_key:
            filename1 = os.path.expanduser('~/.ssh/known_hosts')
            filename2 = _ssh_host_keys_config_dir()
            raise errors.TransportError(
                'Host keys for %s do not match!  %s != %s' %
                (host, our_server_key_hex, server_key_hex),
                ['Try editing {} or {}'.format(filename1, filename2)])

        _paramiko_auth(username, password, host, port, t)
        return t

    def connect_sftp(self, username, password, host, port):
        t = self._connect(username, password, host, port)
        try:
            return t.open_sftp_client()
        except paramiko.SSHException as e:
            self._raise_connection_error(host, port=port, orig_error=e,
                                         msg='Unable to start sftp client')

    def connect_ssh(self, username, password, host, port, command):
        t = self._connect(username, password, host, port)
        try:
            channel = t.open_session()
            cmdline = ' '.join(command)
            channel.exec_command(cmdline)
            return _ParamikoSSHConnection(channel)
        except paramiko.SSHException as e:
            self._raise_connection_error(host, port=port, orig_error=e,
                                         msg='Unable to invoke remote bzr')


_ssh_connection_errors: Tuple[Type[Exception], ...] = (EOFError, OSError, IOError, socket.error)
if paramiko is not None:
    vendor = ParamikoVendor()
    register_ssh_vendor('paramiko', vendor)
    register_ssh_vendor('none', vendor)
    register_default_ssh_vendor(vendor)
    _ssh_connection_errors += (paramiko.SSHException,)
    del vendor


class SubprocessVendor(SSHVendor):
    """Abstract base class for vendors that use pipes to a subprocess."""

    # In general stderr should be inherited from the parent process so prompts
    # are visible on the terminal. This can be overriden to another file for
    # tests, but beware of using PIPE which may hang due to not being read.
    _stderr_target = None

    @staticmethod
    def _check_hostname(arg):
        if arg.startswith('-'):
            raise StrangeHostname(hostname=arg)

    def _connect(self, argv):
        # Attempt to make a socketpair to use as stdin/stdout for the SSH
        # subprocess.  We prefer sockets to pipes because they support
        # non-blocking short reads, allowing us to optimistically read 64k (or
        # whatever) chunks.
        try:
            my_sock, subproc_sock = socket.socketpair()
            osutils.set_fd_cloexec(my_sock)
        except (AttributeError, OSError):
            # This platform doesn't support socketpair(), so just use ordinary
            # pipes instead.
            stdin = stdout = subprocess.PIPE
            my_sock, subproc_sock = None, None
        else:
            stdin = stdout = subproc_sock
        proc = subprocess.Popen(argv, stdin=stdin, stdout=stdout,
                                stderr=self._stderr_target,
                                bufsize=0,
                                **os_specific_subprocess_params())
        if subproc_sock is not None:
            subproc_sock.close()
        return SSHSubprocessConnection(proc, sock=my_sock)

    def connect_sftp(self, username, password, host, port):
        try:
            argv = self._get_vendor_specific_argv(username, host, port,
                                                  subsystem='sftp')
            sock = self._connect(argv)
            return SFTPClient(SocketAsChannelAdapter(sock))
        except _ssh_connection_errors as e:
            self._raise_connection_error(host, port=port, orig_error=e)

    def connect_ssh(self, username, password, host, port, command):
        try:
            argv = self._get_vendor_specific_argv(username, host, port,
                                                  command=command)
            return self._connect(argv)
        except _ssh_connection_errors as e:
            self._raise_connection_error(host, port=port, orig_error=e)

    def _get_vendor_specific_argv(self, username, host, port, subsystem=None,
                                  command=None):
        """Returns the argument list to run the subprocess with.

        Exactly one of 'subsystem' and 'command' must be specified.
        """
        raise NotImplementedError(self._get_vendor_specific_argv)


class OpenSSHSubprocessVendor(SubprocessVendor):
    """SSH vendor that uses the 'ssh' executable from OpenSSH."""

    executable_path = 'ssh'

    def _get_vendor_specific_argv(self, username, host, port, subsystem=None,
                                  command=None):
        args = [self.executable_path,
                '-oForwardX11=no', '-oForwardAgent=no',
                '-oClearAllForwardings=yes',
                '-oNoHostAuthenticationForLocalhost=yes']
        if port is not None:
            args.extend(['-p', str(port)])
        if username is not None:
            args.extend(['-l', username])
        if subsystem is not None:
            args.extend(['-s', '--', host, subsystem])
        else:
            args.extend(['--', host] + command)
        return args


register_ssh_vendor('openssh', OpenSSHSubprocessVendor())


class SSHCorpSubprocessVendor(SubprocessVendor):
    """SSH vendor that uses the 'ssh' executable from SSH Corporation."""

    executable_path = 'ssh'

    def _get_vendor_specific_argv(self, username, host, port, subsystem=None,
                                  command=None):
        self._check_hostname(host)
        args = [self.executable_path, '-x']
        if port is not None:
            args.extend(['-p', str(port)])
        if username is not None:
            args.extend(['-l', username])
        if subsystem is not None:
            args.extend(['-s', subsystem, host])
        else:
            args.extend([host] + command)
        return args


register_ssh_vendor('sshcorp', SSHCorpSubprocessVendor())


class LSHSubprocessVendor(SubprocessVendor):
    """SSH vendor that uses the 'lsh' executable from GNU"""

    executable_path = 'lsh'

    def _get_vendor_specific_argv(self, username, host, port, subsystem=None,
                                  command=None):
        self._check_hostname(host)
        args = [self.executable_path]
        if port is not None:
            args.extend(['-p', str(port)])
        if username is not None:
            args.extend(['-l', username])
        if subsystem is not None:
            args.extend(['--subsystem', subsystem, host])
        else:
            args.extend([host] + command)
        return args


register_ssh_vendor('lsh', LSHSubprocessVendor())


class PLinkSubprocessVendor(SubprocessVendor):
    """SSH vendor that uses the 'plink' executable from Putty."""

    executable_path = 'plink'

    def _get_vendor_specific_argv(self, username, host, port, subsystem=None,
                                  command=None):
        self._check_hostname(host)
        args = [self.executable_path, '-x', '-a', '-ssh', '-2', '-batch']
        if port is not None:
            args.extend(['-P', str(port)])
        if username is not None:
            args.extend(['-l', username])
        if subsystem is not None:
            args.extend(['-s', host, subsystem])
        else:
            args.extend([host] + command)
        return args


register_ssh_vendor('plink', PLinkSubprocessVendor())


def _paramiko_auth(username, password, host, port, paramiko_transport):
    auth = config.AuthenticationConfig()
    # paramiko requires a username, but it might be none if nothing was
    # supplied.  If so, use the local username.
    if username is None:
        username = auth.get_user('ssh', host, port=port,
                                 default=getpass.getuser())
    agent = paramiko.Agent()
    for key in agent.get_keys():
        trace.mutter('Trying SSH agent key %s'
                     % hexlify(key.get_fingerprint()).upper())
        try:
            paramiko_transport.auth_publickey(username, key)
            return
        except paramiko.SSHException as e:
            pass

    # okay, try finding id_rsa or id_dss?  (posix only)
    if _try_pkey_auth(paramiko_transport, paramiko.RSAKey, username, 'id_rsa'):
        return
    if _try_pkey_auth(paramiko_transport, paramiko.DSSKey, username, 'id_dsa'):
        return

    # If we have gotten this far, we are about to try for passwords, do an
    # auth_none check to see if it is even supported.
    supported_auth_types = []
    try:
        # Note that with paramiko <1.7.5 this logs an INFO message:
        #    Authentication type (none) not permitted.
        # So we explicitly disable the logging level for this action
        old_level = paramiko_transport.logger.level
        paramiko_transport.logger.setLevel(logging.WARNING)
        try:
            paramiko_transport.auth_none(username)
        finally:
            paramiko_transport.logger.setLevel(old_level)
    except paramiko.BadAuthenticationType as e:
        # Supported methods are in the exception
        supported_auth_types = e.allowed_types
    except paramiko.SSHException as e:
        # Don't know what happened, but just ignore it
        pass
    # We treat 'keyboard-interactive' and 'password' auth methods identically,
    # because Paramiko's auth_password method will automatically try
    # 'keyboard-interactive' auth (using the password as the response) if
    # 'password' auth is not available.  Apparently some Debian and Gentoo
    # OpenSSH servers require this.
    # XXX: It's possible for a server to require keyboard-interactive auth that
    # requires something other than a single password, but we currently don't
    # support that.
    if ('password' not in supported_auth_types and
            'keyboard-interactive' not in supported_auth_types):
        raise errors.ConnectionError('Unable to authenticate to SSH host as'
                                     '\n  %s@%s\nsupported auth types: %s'
                                     % (username, host, supported_auth_types))

    if password:
        try:
            paramiko_transport.auth_password(username, password)
            return
        except paramiko.SSHException as e:
            pass

    # give up and ask for a password
    password = auth.get_password('ssh', host, username, port=port)
    # get_password can still return None, which means we should not prompt
    if password is not None:
        try:
            paramiko_transport.auth_password(username, password)
        except paramiko.SSHException as e:
            raise errors.ConnectionError(
                'Unable to authenticate to SSH host as'
                '\n  %s@%s\n' % (username, host), e)
    else:
        raise errors.ConnectionError('Unable to authenticate to SSH host as'
                                     '  %s@%s' % (username, host))


def _try_pkey_auth(paramiko_transport, pkey_class, username, filename):
    filename = os.path.expanduser('~/.ssh/' + filename)
    try:
        key = pkey_class.from_private_key_file(filename)
        paramiko_transport.auth_publickey(username, key)
        return True
    except paramiko.PasswordRequiredException:
        password = ui.ui_factory.get_password(
            prompt='SSH %(filename)s password',
            filename=os.fsdecode(filename))
        try:
            key = pkey_class.from_private_key_file(filename, password)
            paramiko_transport.auth_publickey(username, key)
            return True
        except paramiko.SSHException:
            trace.mutter('SSH authentication via %s key failed.'
                         % (os.path.basename(filename),))
    except paramiko.SSHException:
        trace.mutter('SSH authentication via %s key failed.'
                     % (os.path.basename(filename),))
    except OSError:
        pass
    return False


def _ssh_host_keys_config_dir():
    return osutils.pathjoin(bedding.config_dir(), 'ssh_host_keys')


def load_host_keys():
    """
    Load system host keys (probably doesn't work on windows) and any
    "discovered" keys from previous sessions.
    """
    global SYSTEM_HOSTKEYS, BRZ_HOSTKEYS
    try:
        SYSTEM_HOSTKEYS = paramiko.util.load_host_keys(
            os.path.expanduser('~/.ssh/known_hosts'))
    except OSError as e:
        trace.mutter('failed to load system host keys: ' + str(e))
    brz_hostkey_path = _ssh_host_keys_config_dir()
    try:
        BRZ_HOSTKEYS = paramiko.util.load_host_keys(brz_hostkey_path)
    except OSError as e:
        trace.mutter('failed to load brz host keys: ' + str(e))
        save_host_keys()


def save_host_keys():
    """
    Save "discovered" host keys in $(config)/ssh_host_keys/.
    """
    global SYSTEM_HOSTKEYS, BRZ_HOSTKEYS
    bzr_hostkey_path = _ssh_host_keys_config_dir()
    bedding.ensure_config_dir_exists()

    try:
        with open(bzr_hostkey_path, 'w') as f:
            f.write('# SSH host keys collected by bzr\n')
            for hostname, keys in BRZ_HOSTKEYS.items():
                for keytype, key in keys.items():
                    f.write('%s %s %s\n' %
                            (hostname, keytype, key.get_base64()))
    except OSError as e:
        trace.mutter('failed to save bzr host keys: ' + str(e))


def os_specific_subprocess_params():
    """Get O/S specific subprocess parameters."""
    if sys.platform == 'win32':
        # setting the process group and closing fds is not supported on
        # win32
        return {}
    else:
        # We close fds other than the pipes as the child process does not need
        # them to be open.
        #
        # We also set the child process to ignore SIGINT.  Normally the signal
        # would be sent to every process in the foreground process group, but
        # this causes it to be seen only by bzr and not by ssh.  Python will
        # generate a KeyboardInterrupt in bzr, and we will then have a chance
        # to release locks or do other cleanup over ssh before the connection
        # goes away.
        # <https://launchpad.net/products/bzr/+bug/5987>
        #
        # Running it in a separate process group is not good because then it
        # can't get non-echoed input of a password or passphrase.
        # <https://launchpad.net/products/bzr/+bug/40508>
        return {'preexec_fn': _ignore_signals,
                'close_fds': True,
                }


import weakref

_subproc_weakrefs: Set[weakref.ref] = set()


def _close_ssh_proc(proc, sock):
    """Carefully close stdin/stdout and reap the SSH process.

    If the pipes are already closed and/or the process has already been
    wait()ed on, that's ok, and no error is raised.  The goal is to do our best
    to clean up (whether or not a clean up was already tried).
    """
    funcs = []
    for closeable in (proc.stdin, proc.stdout, sock):
        # We expect that either proc (a subprocess.Popen) will have stdin and
        # stdout streams to close, or that we will have been passed a socket to
        # close, with the option not in use being None.
        if closeable is not None:
            funcs.append(closeable.close)
    funcs.append(proc.wait)
    for func in funcs:
        try:
            func()
        except OSError:
            # It's ok for the pipe to already be closed, or the process to
            # already be finished.
            continue


class SSHConnection:
    """Abstract base class for SSH connections."""

    def get_sock_or_pipes(self):
        """Returns a (kind, io_object) pair.

        If kind == 'socket', then io_object is a socket.

        If kind == 'pipes', then io_object is a pair of file-like objects
        (read_from, write_to).
        """
        raise NotImplementedError(self.get_sock_or_pipes)

    def close(self):
        raise NotImplementedError(self.close)


class SSHSubprocessConnection(SSHConnection):
    """A connection to an ssh subprocess via pipes or a socket.

    This class is also socket-like enough to be used with
    SocketAsChannelAdapter (it has 'send' and 'recv' methods).
    """

    def __init__(self, proc, sock=None):
        """Constructor.

        :param proc: a subprocess.Popen
        :param sock: if proc.stdin/out is a socket from a socketpair, then sock
            should breezy's half of that socketpair.  If not passed, proc's
            stdin/out is assumed to be ordinary pipes.
        """
        self.proc = proc
        self._sock = sock
        # Add a weakref to proc that will attempt to do the same as self.close
        # to avoid leaving processes lingering indefinitely.

        def terminate(ref):
            _subproc_weakrefs.remove(ref)
            _close_ssh_proc(proc, sock)
        _subproc_weakrefs.add(weakref.ref(self, terminate))

    def send(self, data):
        if self._sock is not None:
            return self._sock.send(data)
        else:
            return os.write(self.proc.stdin.fileno(), data)

    def recv(self, count):
        if self._sock is not None:
            return self._sock.recv(count)
        else:
            return os.read(self.proc.stdout.fileno(), count)

    def close(self):
        _close_ssh_proc(self.proc, self._sock)

    def get_sock_or_pipes(self):
        if self._sock is not None:
            return 'socket', self._sock
        else:
            return 'pipes', (self.proc.stdout, self.proc.stdin)


class _ParamikoSSHConnection(SSHConnection):
    """An SSH connection via paramiko."""

    def __init__(self, channel):
        self.channel = channel

    def get_sock_or_pipes(self):
        return ('socket', self.channel)

    def close(self):
        return self.channel.close()



# Copyright (C) 2007 by Jelmer Vernooij <jelmer@samba.org>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

"""Bazaar command-line subcommands."""

from ...commands import Command, display_command
from ...errors import (CommandError, ConflictsInTree, NoWorkingTree,
                       UncommittedChanges)
from ...i18n import gettext
from ...option import Option
from ...trace import note
from ...transport import NoSuchFile


def finish_rebase(state, wt, replace_map, replayer):
    from .rebase import rebase
    try:
        # Start executing plan from current Branch.last_revision()
        rebase(wt.branch.repository, replace_map, replayer)
    except ConflictsInTree:
        raise CommandError(gettext(
            "A conflict occurred replaying a commit."
            " Resolve the conflict and run 'brz rebase-continue' or "
            "run 'brz rebase-abort'."))
    # Remove plan file
    state.remove_plan()


class cmd_rebase(Command):
    """Re-base a branch.

    Rebasing is the process of taking a branch and modifying the history so
    that it appears to start from a different point. This can be useful
    to clean up the history before submitting your changes. The tree at the
    end of the process will be the same as if you had merged the other branch,
    but the history will be different.

    The command takes the location of another branch on to which the branch in
    the specified directory (by default, the current working directory)
    will be rebased. If a branch is not specified then the parent branch
    is used, and this is usually the desired result.

    The first step identifies the revisions that are in the current branch that
    are not in the parent branch. The current branch is then set to be at the
    same revision as the target branch, and each revision is replayed on top
    of the branch. At the end of the process it will appear as though your
    current branch was branched off the current last revision of the target.

    Each revision that is replayed may cause conflicts in the tree. If this
    happens the command will stop and allow you to fix them up. Resolve the
    commits as you would for a merge, and then run 'brz resolve' to marked
    them as resolved. Once you have resolved all the conflicts you should
    run 'brz rebase-continue' to continue the rebase operation.

    If conflicts are encountered and you decide that you do not wish to continue
    you can run 'brz rebase-abort'.

    The '--onto' option allows you to specify a different revision in the
    target branch to start at when replaying the revisions. This means that
    you can change the point at which the current branch will appear to be
    branched from when the operation completes.
    """
    takes_args = ['upstream_location?']
    takes_options = [
        'revision', 'merge-type', 'verbose',
        Option(
            'dry-run',
            help="Show what would be done, but don't actually do anything."),
        Option(
            'always-rebase-merges',
            help="Don't skip revisions that merge already present revisions."),
        Option(
            'pending-merges',
            help="Rebase pending merges onto local branch."),
        Option(
            'onto', help='Different revision to replay onto.',
            type=str),
        Option(
            'directory',
            short_name='d',
            help="Branch to replay onto, rather than the one containing the working directory.",
            type=str)
        ]

    @display_command
    def run(self, upstream_location=None, onto=None, revision=None,
            merge_type=None, verbose=False, dry_run=False,
            always_rebase_merges=False, pending_merges=False,
            directory="."):
        from ...branch import Branch
        from ...revisionspec import RevisionSpec
        from ...workingtree import WorkingTree
        from .rebase import (RebaseState1, WorkingTreeRevisionRewriter,
                             generate_simple_plan, rebase, rebase_todo,
                             regenerate_default_revid)
        if revision is not None and pending_merges:
            raise CommandError(gettext(
                "--revision and --pending-merges are mutually exclusive"))

        wt = WorkingTree.open_containing(directory)[0]
        wt.lock_write()
        try:
            state = RebaseState1(wt)
            if upstream_location is None:
                if pending_merges:
                    upstream_location = directory
                else:
                    upstream_location = wt.branch.get_parent()
                    if upstream_location is None:
                        raise CommandError(gettext("No upstream branch specified."))
                    note(gettext("Rebasing on %s"), upstream_location)
            upstream = Branch.open_containing(upstream_location)[0]
            upstream_repository = upstream.repository
            upstream_revision = upstream.last_revision()
            # Abort if there already is a plan file
            if state.has_plan():
                raise CommandError(gettext(
                    "A rebase operation was interrupted. "
                    "Continue using 'brz rebase-continue' or abort using 'brz "
                    "rebase-abort'"))

            start_revid = None
            stop_revid = None
            if revision is not None:
                if len(revision) == 1:
                    if revision[0] is not None:
                        stop_revid = revision[0].as_revision_id(wt.branch)
                elif len(revision) == 2:
                    if revision[0] is not None:
                        start_revid = revision[0].as_revision_id(wt.branch)
                    if revision[1] is not None:
                        stop_revid = revision[1].as_revision_id(wt.branch)
                else:
                    raise CommandError(gettext(
                        "--revision takes only one or two arguments"))

            if pending_merges:
                wt_parents = wt.get_parent_ids()
                if len(wt_parents) in (0, 1):
                    raise CommandError(gettext("No pending merges present."))
                elif len(wt_parents) > 2:
                    raise CommandError(gettext(
                        "Rebasing more than one pending merge not supported"))
                stop_revid = wt_parents[1]
                assert stop_revid is not None, "stop revid invalid"

            # Check for changes in the working tree.
            if (not pending_merges and
                    wt.basis_tree().changes_from(wt).has_changed()):
                raise UncommittedChanges(wt)

            # Pull required revisions
            wt.branch.repository.fetch(upstream_repository, upstream_revision)
            if onto is None:
                onto = upstream.last_revision()
            else:
                rev_spec = RevisionSpec.from_string(onto)
                onto = rev_spec.as_revision_id(upstream)

            wt.branch.repository.fetch(upstream_repository, revision_id=onto)

            if stop_revid is None:
                stop_revid = wt.branch.last_revision()
            repo_graph = wt.branch.repository.get_graph()
            our_new, onto_unique = repo_graph.find_difference(stop_revid, onto)

            if start_revid is None:
                if not onto_unique:
                    self.outf.write(gettext("No revisions to rebase.\n"))
                    return
                if not our_new:
                    self.outf.write(gettext(
                        "Base branch is descendant of current "
                        "branch. Pulling instead.\n"))
                    if not dry_run:
                        wt.pull(upstream, stop_revision=onto)
                    return
            # else: include extra revisions needed to make start_revid mean
            # something.

            # Create plan
            replace_map = generate_simple_plan(
                our_new, start_revid, stop_revid, onto, repo_graph,
                lambda revid, ps: regenerate_default_revid(
                    wt.branch.repository, revid),
                not always_rebase_merges)

            if verbose or dry_run:
                todo = list(rebase_todo(wt.branch.repository, replace_map))
                note(gettext('%d revisions will be rebased:') % len(todo))
                for revid in todo:
                    note("%s" % revid)

            if not dry_run:
                # Write plan file
                state.write_plan(replace_map)

                replayer = WorkingTreeRevisionRewriter(wt, state, merge_type=merge_type)

                finish_rebase(state, wt, replace_map, replayer)
        finally:
            wt.unlock()


class cmd_rebase_abort(Command):
    """Abort an interrupted rebase."""

    takes_options = [
        Option(
            'directory',
            short_name='d',
            help="Branch to replay onto, rather than the one containing the working directory.",
            type=str)]

    @display_command
    def run(self, directory="."):
        from ...workingtree import WorkingTree
        from .rebase import RebaseState1, complete_revert
        wt = WorkingTree.open_containing(directory)[0]
        wt.lock_write()
        try:
            state = RebaseState1(wt)
            # Read plan file and set last revision
            try:
                last_rev_info = state.read_plan()[0]
            except NoSuchFile:
                raise CommandError("No rebase to abort")
            complete_revert(wt, [last_rev_info[1]])
            state.remove_plan()
        finally:
            wt.unlock()


class cmd_rebase_continue(Command):
    """Continue an interrupted rebase after resolving conflicts."""
    takes_options = [
        'merge-type', Option(
            'directory',
            short_name='d',
            help="Branch to replay onto, rather than the one containing the working directory.",
            type=str)]

    @display_command
    def run(self, merge_type=None, directory="."):
        from ...workingtree import WorkingTree
        from .rebase import RebaseState1, WorkingTreeRevisionRewriter
        wt = WorkingTree.open_containing(directory)[0]
        wt.lock_write()
        try:
            state = RebaseState1(wt)
            replayer = WorkingTreeRevisionRewriter(wt, state, merge_type=merge_type)
            # Abort if there are any conflicts
            if len(wt.conflicts()) != 0:
                raise CommandError(gettext(
                    "There are still conflicts present. "
                    "Resolve the conflicts and then run "
                    "'brz resolve' and try again."))
            # Read plan file
            try:
                replace_map = state.read_plan()[1]
            except NoSuchFile:
                raise CommandError(gettext("No rebase to continue"))
            oldrevid = state.read_active_revid()
            if oldrevid is not None:
                oldrev = wt.branch.repository.get_revision(oldrevid)
                replayer.commit_rebase(oldrev, replace_map[oldrevid][0])
            finish_rebase(state, wt, replace_map, replayer)
        finally:
            wt.unlock()


class cmd_rebase_todo(Command):
    """Print list of revisions that still need to be replayed as part of the
    current rebase operation.

    """

    takes_options = [
        Option(
            'directory',
            short_name='d',
            help="Branch to replay onto, rather than the one containing the working directory.",
            type=str)]

    def run(self, directory="."):
        from ...workingtree import WorkingTree
        from .rebase import RebaseState1, rebase_todo
        wt = WorkingTree.open_containing(directory)[0]
        with wt.lock_read():
            state = RebaseState1(wt)
            try:
                replace_map = state.read_plan()[1]
            except NoSuchFile:
                raise CommandError(gettext("No rebase in progress"))
            currentrevid = state.read_active_revid()
            if currentrevid is not None:
                note(gettext("Currently replaying: %s") % currentrevid)
            for revid in rebase_todo(wt.branch.repository, replace_map):
                note(gettext("{0} -> {1}").format(revid, replace_map[revid][0]))


class cmd_replay(Command):
    """Replay commits from another branch on top of this one.

    """

    takes_options = [
        'revision', 'merge-type',
        Option(
            'directory',
            short_name='d',
            help="Branch to replay onto, rather than the one containing the working directory.",
            type=str)]
    takes_args = ['location']
    hidden = True

    def run(self, location, revision=None, merge_type=None, directory="."):
        from ... import ui
        from ...branch import Branch
        from ...workingtree import WorkingTree
        from .rebase import (RebaseState1, WorkingTreeRevisionRewriter,
                             regenerate_default_revid)

        from_branch = Branch.open_containing(location)[0]

        if revision is not None:
            if len(revision) == 1:
                if revision[0] is not None:
                    todo = [revision[0].as_revision_id(from_branch)]
            elif len(revision) == 2:
                from_revno, from_revid = revision[0].in_history(from_branch)
                to_revno, to_revid = revision[1].in_history(from_branch)
                if to_revid is None:
                    to_revno = from_branch.revno()
                todo = []
                for revno in range(from_revno, to_revno + 1):
                    todo.append(from_branch.get_rev_id(revno))
            else:
                raise CommandError(gettext(
                    "--revision takes only one or two arguments"))
        else:
            raise CommandError(gettext("--revision is mandatory"))

        wt = WorkingTree.open(directory)
        wt.lock_write()
        try:
            state = RebaseState1(wt)
            replayer = WorkingTreeRevisionRewriter(wt, state, merge_type=merge_type)
            pb = ui.ui_factory.nested_progress_bar()
            try:
                for revid in todo:
                    pb.update(gettext("replaying commits"), todo.index(revid), len(todo))
                    wt.branch.repository.fetch(from_branch.repository, revid)
                    newrevid = regenerate_default_revid(wt.branch.repository, revid)
                    replayer(revid, newrevid, [wt.last_revision()])
            finally:
                pb.finished()
        finally:
            wt.unlock()


class cmd_pseudonyms(Command):
    """Show a list of 'pseudonym' revisions.

    Pseudonym revisions are revisions that are roughly the same revision,
    usually because they were converted from the same revision in a
    foreign version control system.
    """

    takes_args = ['repository?']
    hidden = True

    def run(self, repository=None):
        from ...controldir import ControlDir
        dir, _ = ControlDir.open_containing(repository)
        r = dir.find_repository()
        from .pseudonyms import find_pseudonyms
        for pseudonyms in find_pseudonyms(r, r.all_revision_ids()):
            self.outf.write(", ".join(pseudonyms) + "\n")


class cmd_rebase_foreign(Command):
    """Rebase revisions based on a branch created with a different import tool.

    This will change the identity of revisions whose parents
    were mapped from revisions in the other version control system.

    You are recommended to run "brz check" in the local repository
    after running this command.
    """
    takes_args = ['new_base?']
    takes_options = [
        'verbose',
        Option("idmap-file", help="Write map with old and new revision ids.",
               type=str),
        Option(
            'directory',
            short_name='d',
            help="Branch to replay onto, rather than the one containing the working directory.",
            type=str)
        ]

    def run(self, new_base=None, verbose=False, idmap_file=None, directory="."):
        from ... import urlutils
        from ...branch import Branch
        from ...foreign import update_workingtree_fileids
        from ...workingtree import WorkingTree
        from .pseudonyms import (find_pseudonyms,
                                 generate_rebase_map_from_pseudonyms,
                                 pseudonyms_as_dict)
        from .upgrade import create_deterministic_revid, upgrade_branch

        try:
            wt_to = WorkingTree.open(directory)
            branch_to = wt_to.branch
        except NoWorkingTree:
            wt_to = None
            branch_to = Branch.open(directory)

        stored_loc = branch_to.get_parent()
        if new_base is None:
            if stored_loc is None:
                raise CommandError(gettext(
                    "No pull location known or specified."))
            else:
                display_url = urlutils.unescape_for_display(
                    stored_loc, self.outf.encoding)
                self.outf.write(gettext("Using saved location: %s\n") % display_url)
                new_base = Branch.open(stored_loc)
        else:
            new_base = Branch.open(new_base)

        branch_to.repository.fetch(
            new_base.repository, revision_id=branch_to.last_revision())

        pseudonyms = pseudonyms_as_dict(find_pseudonyms(
            branch_to.repository, branch_to.repository.all_revision_ids()))

        def generate_rebase_map(revision_id):
            return generate_rebase_map_from_pseudonyms(
                pseudonyms, branch_to.repository.get_ancestry(revision_id),
                branch_to.repository.get_ancestry(new_base.last_revision()))
        def determine_new_revid(old_revid, new_parents):
            return create_deterministic_revid(old_revid, new_parents)
        branch_to.lock_write()
        try:
            graph = branch_to.repository.get_graph()
            renames = upgrade_branch(
                branch_to, generate_rebase_map,
                determine_new_revid, allow_changes=True,
                verbose=verbose)
            if wt_to is not None:
                basis_tree = wt_to.basis_tree()
                basis_tree.lock_read()
                try:
                    update_workingtree_fileids(wt_to, basis_tree)
                finally:
                    basis_tree.unlock()
        finally:
            branch_to.unlock()

        if renames == {}:
            note(gettext("Nothing to do."))

        if idmap_file is not None:
            f = open(idmap_file, 'w')
            try:
                for oldid, newid in renames.iteritems():
                    f.write("{}\t{}\n".format(oldid, newid))
            finally:
                f.close()

        if wt_to is not None:
            wt_to.set_last_revision(branch_to.last_revision())
